from random import random
import numpy as np
import scipy.special
from numpy.random import binomial, multivariate_normal, normal, uniform
import pandas as pd

from itertools import product

from sklearn.model_selection import train_test_split


def eta_sample(n):
    return np.random.uniform(-1, 1, size=n)


def epsilon_sample(n):
    return np.random.uniform(-1, 1, size=n)


def exp_te(x):
    return np.exp(2*x[0])


def ln_te(x):
    return np.log(1+x[0])


def build_data_frame(
    confounder_n,
    covariate_n,
    w, v, y, x, **kwargs,
):
    data_dict = {}
    for i in range(confounder_n):
        data_dict[f'w_{i}'] = w[:, i]

    for i in range(covariate_n):
        data_dict[f'c_{i}'] = v[:, i]

    def multi_var(x, name):
        if len(x.shape) == 1 or x.shape[1] == 1:
            data_dict[name] = x
        else:
            for i in range(x.shape[1]):
                data_dict[f'{name}_{i}'] = x[:, i]

    multi_var(x, 'treatment')
    multi_var(y, 'outcome')

    for k, v in kwargs.items():
        data_dict[k] = v

    data = pd.DataFrame(data_dict)
    train, val = train_test_split(data)
    
    return train, val


def coupon_dataset(n_users, treatment_style='binary', with_income=False):
    if with_income:
        income = np.random.normal(500, scale=15, size=n_users)
        gender = np.random.randint(0, 2, size=n_users)
        coupon = gender * 20 + 110 + income / 50 \
            + np.random.normal(scale=5, size=n_users)
        if treatment_style == 'binary':
            coupon = (coupon > 120).astype(int)
        amount = coupon * 150 + gender * 100 + 150 \
            + income / 5 + np.random.normal(size=n_users)
        time_spent = coupon * 10 + amount / 10

        df = pd.DataFrame({
            'gender': gender,
            'coupon': coupon,
            'amount': amount,
            'income': income,
            'time_spent': time_spent,
        })
    else:
        gender = np.random.randint(0, 2, size=n_users)
        coupon = gender * 20 + 150 + np.random.normal(scale=5, size=n_users)
        if treatment_style == 'binary':
            coupon = (coupon > 150).astype(int)
        amount = coupon * 30 + gender * 100 \
            + 150 + np.random.normal(size=n_users)
        time_spent = coupon * 100 + amount / 10

        df = pd.DataFrame({
            'gender': gender,
            'coupon': coupon,
            'amount': amount,
            'time_spent': time_spent,
        })

    return df

def generate_controls_outcome(d):
    beta = uniform(-3, 3, d)
    return lambda x: np.dot(x, beta) + normal(0, 1)

def binary_data(n=2000, d=5, n_test=250):
    def _binary_data(n, d, ctrl, ce, propensity):
        np.random.seed(2022)
        # Generate covariates
        v = multivariate_normal(np.zeros(d), np.diag(np.ones(d)), n)
        # Generate treatment
        x = np.apply_along_axis(lambda x: binomial(1, propensity(x), 1)[0], 1, v)
        # Calculate outcome
        y0 = np.apply_along_axis(lambda x: ctrl(x), 1, v)
        treat_effect = np.apply_along_axis(lambda x: ce(x), 1, v)
        y = y0 + treat_effect * x
        return (y, x, v)

    treatment_effect = lambda x: (1 if x[1] > 0.1 else 0)*8
    propensity = lambda x: (0.8 if (x[2]>-0.5 and x[2]<0.5) else 0.2)
    controls_outcome = generate_controls_outcome(d)
    
    return _binary_data(n, d, controls_outcome, treatment_effect, propensity)

def sq_data(n, d, n_x, random_seed=123):
    np.random.seed(random_seed)
    true_te = lambda X: np.hstack([X[:, [0]]**2 + 1, np.ones((X.shape[0], n_x - 1))])
    v = np.random.normal(0, 1, size=(n, d))
    x = np.random.normal(0, 1, size=(n, n_x))
    for t in range(n_x):
        x[:, t] = np.random.binomial(1, scipy.special.expit(v[:, 0]))
    y = np.sum(true_te(v) * x, axis=1, keepdims=True) + np.random.normal(0, .5, size=(n, 1))
    return y, x, v

def multi_continuous_treatment(
    n=6000,
    n_w=30,
    n_v=5,
    random_seed=2022,
):
    np.random.seed(random_seed)

    support_size = 5
    support_Y = np.random.choice(
        np.arange(n_w), size=support_size, replace=False
    )

    coefs_Y = np.random.uniform(0, 1, size=support_size)
    support_T = support_Y
    coefs_T = np.random.uniform(0, 1, size=support_size)

    W = np.random.normal(0, 1, size=(n, n_w))
    X = np.random.uniform(0, 1, size=(n, n_v))

    TE1 = np.array([exp_te(x_i) for x_i in X])
    TE2 = np.array([ln_te(x_i) for x_i in X]).flatten()
    T = np.dot(W[:, support_T], coefs_T) + eta_sample(n)
    Y = TE1 * T + TE2 * T**2 + \
        np.dot(W[:, support_Y], coefs_Y) + epsilon_sample(n)
    T = T.reshape(-1, 1)
    x = np.concatenate((T, T**2), axis=1)

    train, val = build_data_frame(confounder_n=n_w,
                                  covariate_n=n_v,
                                  w=W, v=X, y=Y, x=x, te1=TE1, te2=TE2)
    # test data
    X_test = np.random.uniform(0, 1, size=(100, n_v))
    X_test[:, 0] = np.linspace(0, 1, 100)

    data_test_dic = {}
    for i in range(n_v):
        data_test_dic[f'c_{i}'] = X_test[:, i]
    data_test = pd.DataFrame(data_test_dic)

    expected_te1 = np.array([exp_te(x_i) for x_i in X_test])
    expected_te2 = np.array([ln_te(x_i) for x_i in X_test]).flatten()

    return train, val, (data_test, expected_te1, expected_te2)


def single_binary_treatment(
    n=1000,
    confounder_n=30,
    covariate_n=4,
    random_seed=2022,
):
    np.random.seed(random_seed)
    support_size = 5
    # Outcome support
    support_Y = np.random.choice(
        range(confounder_n), size=support_size, replace=False)
    coefs_Y = np.random.uniform(0, 1, size=support_size)
    # Treatment support
    support_T = support_Y
    coefs_T = np.random.uniform(0, 1, size=support_size)

    # Generate controls, covariates, treatments and outcomes
    W = np.random.normal(0, 1, size=(n, confounder_n))
    V = np.random.uniform(0, 1, size=(n, covariate_n))
    # Heterogeneous treatment effects
    TE = np.array([exp_te(x_i) for x_i in V])
    # Define treatment
    log_odds = np.dot(W[:, support_T], coefs_T) + eta_sample(n)
    T_sigmoid = 1/(1 + np.exp(-log_odds))
    x = np.array([np.random.binomial(1, p) for p in T_sigmoid])
    # Define the outcome
    Y = TE * x + np.dot(W[:, support_Y], coefs_Y) + epsilon_sample(n)
    train, val = build_data_frame(confounder_n=confounder_n,
                                  covariate_n=covariate_n,
                                  w=W, v=V, y=Y, x=x, TE=TE)

    return train, val, TE


def single_continuous_treatment(num=2000,
                                confounder_n=30,
                                covariate_n=1,
                                random_seed=2022,
                                data_frame=True):
    np.random.seed(random_seed)
    support_size = 5
    support_y = np.random.choice(
        np.arange(confounder_n), size=support_size, replace=False
    )
    coefs_y = np.random.uniform(0, 1, size=support_size)
    support_t = support_y
    coefs_t = np.random.uniform(0, 1, size=support_size)
    w = np.random.normal(0, 1, size=(num, confounder_n))
    c = np.random.uniform(0, 1, size=(num, covariate_n))
    TE = np.array([exp_te(ci) for ci in c])
    x = np.dot(w[:, support_t], coefs_t) + eta_sample(num)
    y = TE * x + np.dot(w[:, support_y], coefs_y)\
        + epsilon_sample(num)

    x_test = np.array(list(product(np.arange(0, 1, 0.01), repeat=covariate_n)))
    if data_frame:
        train, val = build_data_frame(confounder_n=confounder_n,
                                      covariate_n=covariate_n,
                                      w=w, v=c, y=y, x=x, TE=TE)
        return train, val, TE


def meaningless_discrete_dataset_(num, treatment_effct,
                                  confounder_n=2,
                                  w_var=5,
                                  eps=1e-4,
                                  data_frame=True,
                                  random_seed=2022,
                                  instrument=False):
    """Generate a dataset where the treatment and outcome have some
    confounders while the relation between the treatment and outcome
    is linear. The treatment is an array of integers where each integer
    indicates the treatment group assigned to the corresponding example.
    The outcome is an array of float, i.e., we are building continuous
    outcome.

    Parameters
    ----------
    num : int
        The number of examples in the dataset.
    confounder_n : int
        The number of confounders of the treatment and outcome.
    treatment_effct : list, optional. Defaults to None.
    w_var : float, optional. Defaults to 0.5.
        Variance of the confounder around its mean.
    eps : float, optional. Defaults to 1e-4.
        Noise level imposed to the data generating process.
    data_frame : bool, optional. Defaults to True.
        Return pandas.DataFrame if True.
    random_seed : int, optional. Defaults to 2022.
    instrument : bool, optional. Defaults to False.
        Add instrument variables to the dataset if True.

    Returns
    ----------
    pandas.DataFrame, optional.
        w_j's are confounders of outcome and treatment.
    """
    np.random.seed(random_seed)

    # Build treatment x which depends on the confounder w
    x_num = len(treatment_effct)
    w = [
        np.random.normal(0, w_var*np.random.random_sample(), size=(num, 1))
        for i in range(confounder_n)
    ]
    w = np.concatenate(tuple(w), axis=1)
    w_coef = np.random.rand(x_num, confounder_n)
    x = w.dot(w_coef.T) + np.random.normal(0, eps, size=(num, x_num))
    if instrument:
        z = None
    x = x.argmax(axis=1)
    x_one_hot = np.eye(x_num)[x]

    # Now we build the outcome y which depends on both x and w
    x_coef = np.random.randn(1, confounder_n)
    x_coef = np.concatenate(
        (np.array(treatment_effct).reshape(1, -1), x_coef), axis=1
    )
    x_ = np.concatenate((x_one_hot, w), axis=1)
    y = x_.dot(x_coef.T) + np.random.normal(0, eps, size=(num, 1))

    # Return the dataset
    if data_frame:
        data_dict = {}
        data_dict['treatment'] = x
        if instrument:
            data_dict['instrument'] = z
        for i, j in enumerate(w.T):
            data_dict[f'w_{i}'] = j
        data_dict['outcome'] = y.reshape(num,)
        data = pd.DataFrame(data_dict)
        return data
    else:
        if instrument:
            return (x, w, z, y)
        else:
            return (x, w, y)



def meaningless_discrete_dataset(num, confounder_n,
                                 treatment_effct=None,
                                 prob=None,
                                 w_var=0.5,
                                 eps=1e-4,
                                 coef_range=5e4,
                                 data_frame=True,
                                 random_seed=2022):
    np.random.seed(random_seed)
    samples = np.random.multinomial(num, prob)
    # build treatment x with shape (num,), where the number of types
    # of treatments is len(prob) and each treatment i is assigned a
    # probability prob[i]
    x = []
    for i, sample in enumerate(samples):
        x += [i for j in range(sample)]
    np.random.shuffle(x)

    # construct the confounder w
    w = [
        np.random.normal(0, w_var, size=(num,)) for i in range(confounder_n)
    ]
    for i, w_ in enumerate(w, 1):
        x = x + w_
    x = np.round(x).astype(int)
    for i, j in enumerate(x):
        if j > len(prob) - 1:
            x[i] = len(prob) - 1
        elif j < 0:
            x[i] = 0

    # construct the outcome y
    coef = np.random.randint(int(coef_range*eps), size=(confounder_n,))
    y = np.random.normal(eps, size=(num,))
    for i in range(len(y)):
        y[i] = y[i] + treatment_effct[x[i]] * x[i]
    for i, j in zip(coef, w):
        y += i * j

    if data_frame:
        data_dict = {}
        data_dict['treatment'] = x
        for i, j in enumerate(w):
            data_dict[f'w_{i}'] = j
        data_dict['outcome'] = y
        data = pd.DataFrame(data_dict)
        return data, coef
    else:
        return (x, w, y, coef)



import numpy as np

from sklearn import clone
from sklearn.preprocessing import OrdinalEncoder, OneHotEncoder
from collections import defaultdict

from .base_models import BaseEstModel
from .utils import (
    convert2array,
    get_groups,
    get_treat_control,
    get_wv,
    cartesian,
    get_tr_ctrl,
)


class SLearner(BaseEstModel):
    r"""
    SLearn uses one machine learning model to compute the causal effects.
    Specifically, we fit a model to predict outcome (y) from treatment (x) and
    adjustment (w):

    .. math::

        y = f(x, w).

    The causal effect is then calculated as

    .. math::

        \text{causal_effect} = f(x=1, w) - f(x=0, w).

    Attributes
    ----------
    model : estimator, optional
        The machine learning model for the control group data. Any valid x0_model
        should implement the fit() and predict() methods.

    random_state : int. Defaults to 2022.

    _is_fitted : bool. Defaults to False.
        True if the TLearner is fitted ortherwise False.

    _y_d : int

    treatment : list of str, optional
        Names of the treatments.

    outcome : list of str, optional
        Names of the outcomes.

    adjustment : list of str, optional
        Names of the adjustment set.

    covariate : list of str, optional
        Names of the covariates.

    combined_treat : bool
        Whether use the combined treat technique when training the model.

    categories : str, optional. Defaults to 'auto'.

    combined_treat : bool. Defaults to True.
        Whether combine multiple treatments into a single treatment.

    _is_fitted : bool
        True if the model is fitted ortherwise False.

    Methods
    ----------
    fit(data, outcome, treatment, adjustment, covariate, treat, control, combined_treatment)

    estimate(data, quantity=None)

    _fit_combined_treat(x, wv, y, treat, control, categories, **kwargs)
        Fit function when combined_treat is set to True.

    _comp_transformer(x, categories='auto')
        Transform the discrete treatment into one-hot vectors when combined_treat
        is set to True.

    _fit_separate_treat(x, wv, y, categories)
        Fit function when combined_treat is set to False.

    _prepare4est(data, outcome, treatment, adjustment, individual=None)
        Prepare (fit the model) for estimating various quantities including
        ATE, CATE, ITE, and CITE.

    _prepare_combined_treat(wv)

    _prepare_separate_treat(wv)
    """

    def __init__(
        self,
        model,
        random_state=2022,
        is_discrete_treatment=True,
        categories="auto",
        *args,
        **kwargs,
    ):
        """
        Parameters
        ----------
        model : estimator, optional
            The base machine learning model for training SLearner. Any model
            should be some valid machine learning model with fit() and
            predict() functions.

        random_state : int

        categories : str
        """
        assert is_discrete_treatment is True

        self.model = clone(model)
        self._is_fitted = False

        super().__init__(
            random_state=random_state,
            is_discrete_treatment=is_discrete_treatment,
            categories=categories,
            *args,
            **kwargs,
        )

    def fit(
        self,
        data,
        outcome,
        treatment,
        adjustment=None,
        covariate=None,
        treat=None,
        control=None,
        combined_treatment=True,
        **kwargs,
    ):
        """Fit the SLearner in the dataset.

        Parameters
        ----------
        data : pandas.DataFrame
            Training dataset for training the estimator.

        outcome : list of str, optional.
            List of names of the outcome features.

        treatment : list of str, optional
            List of names of the treatment features.

        adjustment : List of str, optional
            Lisf ot names of adjustment set ensuring the unconfoundness, by default None

        covariate : List of str, optional
            Covariate features, by default None

        treat : int, optional
            Label of the intended treatment group, by default None

        control : int, optional
            Label of the intended control group, by default None

        combined_treatment : bool, optional
            Only modify this parameter for multiple treatments, where multiple discrete
            treatments are combined to give a single new group of discrete treatment if
            set as True.
            When combined_treatment is set to True, then if there are multiple
            treatments, we can use the combined_treatment technique to covert
            the multiple discrete classification tasks into a single discrete
            classification task. For an example, if there are two different
            binary treatments:
                treatment_1: x_1 | x_1 \in {'sleep', 'run'},
                treatment_2: x_2 | x_2 \in {'study', 'work'},
            then we can convert these two binary classification tasks into
            a single classification task with 4 different classes:
                treatment: x | x \in {0, 1, 2, 3},
            where, for example, 1 stands for ('sleep' and 'stuy').

        Returns
        -------
        instance of SLearner
        """
        assert (
            adjustment is not None or covariate is not None
        ), "Need adjustment set or covariates to perform estimation."

        super().fit(
            data,
            outcome,
            treatment,
            adjustment=adjustment,
            covariate=covariate,
        )

        self.combined_treat = combined_treatment

        # get numpy data
        y, x, w, v = convert2array(data, outcome, treatment, adjustment, covariate)
        self._y_d = y.shape[1]

        # get categories for treatment transformer
        if self.categories == "auto" or self.categories is None:
            categories = "auto"
        else:
            categories = list(self.categories)

        # self.transformer = OneHotEncoder(categories=categories)
        wv = get_wv(w, v)
        self._w = w
        self._wv = wv

        if combined_treatment:
            return self._fit_combined_treat(
                x, wv, y, treat, control, categories, **kwargs
            )
        else:
            return self._fit_separate_treat(x, wv, y, categories, **kwargs)

    def _prepare4est(self, data=None, **kwargs):
        if not self._is_fitted:
            raise Exception("The estimator has not been fitted yet.")

        if data is None:
            wv = self._wv
        else:
            w, v = convert2array(data, self.adjustment, self.covariate)
            # if w is None:
            #     w = self._w

            wv = get_wv(w, v)

        if self.combined_treat:
            return self._prepare_combined_treat(wv)
        else:
            return self._prepare_separate_treat(wv)

    def estimate(self, data=None, quantity=None, **kwargs):
        """Estimate the causal effect with the type of the quantity.

        Parameters
        ----------
        data : pandas.DataFrame, optional. Defaults to None.
            If None, then the training data is used as data.

        quantity : str, optional. Defaults to None
            Option for returned estimation result. The possible values of quantity include:

                1. *'CATE'* : the estimator will evaluate the CATE;

                2. *'ATE'* : the estimator will evaluate the ATE;

                3. *None* : the estimator will evaluate the ITE or CITE.

        Returns
        -------
        ndarray
        """
        effect = self._prepare4est(data, **kwargs)
        if quantity == "CATE" or quantity == "ATE":
            return np.mean(effect, axis=0)
        else:
            return effect

    def effect_nji(self, data=None):
        y_nji = self._prepare4est(data=data)

        if y_nji.ndim == 3:
            n, y_d, x_d = y_nji.shape
        elif y_nji.ndim == 2:
            n, y_d = y_nji.shape
            x_d = 1
        else:
            n, y_d, x_d = y_nji.shape[0], 1, 1

        y_nji = y_nji.reshape(n, y_d, x_d)
        zeros_ = np.zeros((n, y_d, 1))
        y_nji = np.concatenate((zeros_, y_nji), axis=2)

        return y_nji

    def _fit_combined_treat(self, x, wv, y, treat, control, categories, **kwargs):
        """Fit function which is used when multiple treatments are combined to
        give a single equivalent treatment vector.

        Parameters
        ----------
        x : np.array
            Treatment variables with shape (n, x_d)

        wv : np.array
            Covariate variables with shape (n, wv_d)

        y : np.array
            Outcome vevariablesctor with shape (n, y_d)

        treat : int or list, optional
            If there is only one treament, then treat indicates the treatment
            group. If there are multiple treatment groups, then treat should
            be a list of str with length equal to the number of treatments.
            For example, when there are multiple
            discrete treatments, array(['run', 'read']) means the treat value of
            the first treatment is taken as 'run' and that of the second treatment
            is taken as 'read'.

        control : int or list, optional
            See treat for more information

        categories : str

        Returns
        -------
        instance of SLearner
        """
        # Converting treatment to array with shape (n, num_treatments)
        # self.transformer = OrdinalEncoder(categories=categories)
        # self.transformer.fit(x)
        # x = self.transformer.transform(x)

        # # For multiple treatments: divide the data into several groups
        # group_categories = self.transformer.categories_
        # n_treatments = len(group_categories)

        x = self._comp_transformer(x, categories=categories)

        self._x_d = x.shape[1]
        self.treat = treat
        self.control = control

        x = np.concatenate((wv, x), axis=1)
        y = y.squeeze()

        self.model.fit(x, y, **kwargs)

        self._is_fitted = True

        return self

    def _comp_transformer(self, x, categories="auto"):
        """Transform the discrete treatment into one-hot vectors.

        Parameters
        ----------
        x : ndarray, shape (n, x_d)
            An array containing the information of the treatment variables

        categories : str or list, optional
            by default 'auto'

        Returns
        -------
        ndarray
            The transformed one-hot vectors
        """
        if x.shape[1] > 1:
            if not self._is_fitted:
                self.ord_transformer = OrdinalEncoder(categories=categories)
                self.ord_transformer.fit(x)

                labels = [np.arange(len(c)) for c in self.ord_transformer.categories_]
                labels = cartesian(labels)
                categories = [np.arange(len(labels))]

                self.label_dict = {tuple(k): i for i, k in enumerate(labels)}

            x_transformed = self.ord_transformer.transform(x).astype(int)
            x = np.full((x.shape[0], 1), np.NaN)

            for i, x_i in enumerate(x_transformed):
                x[i] = self.label_dict[tuple(x_i)]

        if not self._is_fitted:
            self.oh_transformer = OneHotEncoder(categories=categories)
            self.oh_transformer.fit(x)

        x = self.oh_transformer.transform(x).toarray()

        return x

    def _fit_separate_treat(self, x, wv, y, categories, **kwargs):
        """Fit function which is used when multiple treatments are treated separately,
        i.e., when combined_treat is False.

        Parameters
        ----------
        x : ndarray
            Treatment variables.

        wv : ndarray
            Covariate variables.

        y : ndarray
            Outcome variables.

        categories : str, optional. Defaults to 'auto'.

        Returns
        -------
        instance of SLearner
        """
        self.transformer = OneHotEncoder(categories=categories)
        self.transformer.fit(x)
        x = self.transformer.transform(x).toarray()
        self._x_d = x.shape[1]
        x = np.concatenate((wv, x), axis=1)
        y = y.squeeze()

        # self.model.fit(x, y, *args, **kwargs)
        self.model.fit(x, y, **kwargs)

        self._is_fitted = True

        return self

    def _prepare_combined_treat(self, wv):
        n = wv.shape[0]
        self.treat = get_tr_ctrl(
            self.treat,
            self._comp_transformer,
            treat=True,
            one_hot=False,
            discrete_treat=True,
        )
        self.control = get_tr_ctrl(
            self.control,
            self._comp_transformer,
            treat=False,
            one_hot=False,
            discrete_treat=True,
        )

        treat = np.zeros(
            self._x_d,
        )
        control = np.zeros(
            self._x_d,
        )

        treat[self.treat] = 1
        control[self.control] = 1

        xt = np.repeat(treat.reshape(1, -1), n, axis=0)
        x0 = np.repeat(control.reshape(1, -1), n, axis=0)

        if len(xt.shape) == 1:
            xt = xt.reshape(-1, 1)
            x0 = x0.reshape(-1, 1)

        xt = np.concatenate((wv, xt), axis=1)
        x0 = np.concatenate((wv, x0), axis=1)

        yt = self.model.predict(xt)
        y0 = self.model.predict(x0)
        return yt - y0

    def _prepare_separate_treat(self, wv):
        n = wv.shape[0]
        x_control = np.zeros((1, self._x_d))
        x_control[:, 0] = 1
        x_control = np.repeat(x_control, n, axis=0).astype(int)
        x_control = np.concatenate((wv, x_control), axis=1)

        f_nji = np.full((n, self._y_d, self._x_d - 1), np.NaN)
        f_nj0 = self.model.predict(x_control)

        for i in range(self._x_d - 1):
            x_treat = np.zeros((1, self._x_d))
            x_treat[:, i + 1] = 1
            x_treat = np.repeat(x_treat, n, axis=0).astype(int)
            x_treat = np.concatenate((wv, x_treat), axis=1)
            fnji = (self.model.predict(x_treat) - f_nj0).reshape(n, self._y_d)
            f_nji[:, :, i] = fnji

        return f_nji.squeeze()

    # def __repr__(self) -> str:
    #     return f'SLearner'


class TLearner(BaseEstModel):
    """
    TLearner uses two machine learning models to estimate the causal
    effect. Specifically, we
    1. fit two models for the treatment group (x=treat) and control group
    (x=control), respectively:
        y1 = x1_model(w) with data where x=treat,
        y0 = x0_model(w) with data where x=control;
    2. compute the causal effect as the difference between these two models:
        causal_effect = x1_model(w) - x0_model(w).

    Attributes
    -----------
    xt_model : estimator, optional
        The machine learning model for the treatment group data. Any valid xt_model
        should implement the fit() and predict() methods.

    x0_model : estimator, optional
        The machine learning model for the control group data. Any valid x0_model
        should implement the fit() and predict() methods.

    random_state : int. Defaults to 2022.

    _is_fitted : bool. Defaults to False.
        True if the TLearner is fitted ortherwise False.

    _y_d : int

    transformer : OrdinalEncoder

    categories : str

    treatment : list of str, optional
        Names of the treatments.

    outcome : list of str, optional
        Names of the outcomes.

    adjustment : list of str, optional
        Names of the adjustment set.

    covariate : list of str, optional
        Names of the covariates.

    combined_treat : bool
        Whether use the combined treat technique when training the model.

    _wv : ndarray with shape (n, w_d + v_d)

    _w : ndarray with shape (n, w_d)

    Methods
    ----------
    fit(data, outcome, treatment, adjustment, covariate, treat, control, combined_treatment)

    estimate(data, quantity=None)

    _fit_combined_treat(x, wv, y, treat, control, categories, **kwargs)
        Fit function when combined_treat is set to True.

    _comp_transformer(x, categories='auto')
        Transform the discrete treatment into one-hot vectors when combined_treat
        is set to True.

    _fit_separate_treat(x, wv, y, categories)
        Fit function when combined_treat is set to False.

    _prepare4est(data, outcome, treatment, adjustment, individual=None)
        Prepare (fit the model) for estimating various quantities including
        ATE, CATE, ITE, and CITE.

    _prepare_combined_treat(wv)

    _prepare_separate_treat(wv)
    """

    def __init__(
        self,
        model,
        random_state=2022,
        is_discrete_treatment=True,
        categories="auto",
        **kwargs,
    ):
        """

        Parameters
        ----------
        model : estimator, optional
            The base machine learning model for training TLearner. Any model
            should be some valid machine learning model with fit() and
            predict() functions.

        random_state : int

        categories : str
        """
        assert is_discrete_treatment is True

        self.model = clone(model)
        self.xt_model = clone(model)
        self.x0_model = clone(model)
        self._is_fitted = False

        super().__init__(
            random_state=random_state,
            is_discrete_treatment=is_discrete_treatment,
            categories=categories,
            **kwargs,
        )

    def fit(
        self,
        data,
        outcome,
        treatment,
        adjustment=None,
        covariate=None,
        treat=None,
        control=None,
        combined_treatment=True,
        **kwargs,
    ):
        """Fit the TLearner in the dataset.

        Parameters
        ----------
        data : pandas.DataFrame
            Training dataset for training the estimator.

        outcome : list of str, optional.
            List of names of the outcome features.

        treatment : list of str, optional
            List of names of the treatment features.

        adjustment : List of str, optional
            Lisf ot names of adjustment set ensuring the unconfoundness, by default None

        covariate : List of str, optional
            Covariate features, by default None

        treat : int, optional
            Label of the intended treatment group, by default None

        control : int, optional
            Label of the intended control group, by default None

        combined_treatment : bool, optional
            Only modify this parameter for multiple treatments, where multiple discrete
            treatments are combined to give a single new group of discrete treatment if
            set as True.
            When combined_treatment is set to True, then if there are multiple
            treatments, we can use the combined_treatment technique to covert
            the multiple discrete classification tasks into a single discrete
            classification task. For an example, if there are two different
            binary treatments:
                treatment_1: x_1 | x_1 \in {'sleep', 'run'},
                treatment_2: x_2 | x_2 \in {'study', 'work'},
            then we can convert these two binary classification tasks into
            a single classification task with 4 different classes:
                treatment: x | x \in {0, 1, 2, 3},
            where, for example, 1 stands for ('sleep' and 'stuy').

        Returns
        -------
        instance of TLearner
        """
        assert (
            adjustment is not None or covariate is not None
        ), "Need adjustment set or covariates to perform estimation."
        super().fit(
            data,
            outcome,
            treatment,
            adjustment=adjustment,
            covariate=covariate,
        )

        self.combined_treat = combined_treatment

        y, x, w, v = convert2array(data, outcome, treatment, adjustment, covariate)
        self._y_d = y.shape[1]

        if self.categories == "auto" or self.categories is None:
            categories = "auto"
        else:
            categories = list(self.categories)

        # self.transformer = OrdinalEncoder(categories=categories)
        # self.transformer.fit(x)
        # x = self.transformer.transform(x)

        # self.group_categories = self.transformer.categories_
        wv = get_wv(w, v)
        self._wv = wv
        self._w = w

        if combined_treatment:
            return self._fit_combined_treat(
                x, wv, y, treat, control, categories, **kwargs
            )
        else:
            return self._fit_separate_treat(x, wv, y, categories, **kwargs)

    def estimate(self, data=None, quantity=None, **kwargs):
        """Estimate the causal effect with the type of the quantity.

        Parameters
        ----------
        data : pandas.DataFrame, optional. Defaults to None.
            If None, then the training data is used as data.

        quantity : str, optional. Defaults to None
            The possible values of quantity include:
                'CATE' : the estimator will evaluate the CATE;
                'ATE' : the estimator will evaluate the ATE;
                None : the estimator will evaluate the ITE or CITE.

        Returns
        -------
        ndarray
        """
        effect = self._prepare4est(data)
        if quantity == "CATE" or quantity == "ATE":
            return np.mean(effect, axis=0)
        else:
            return effect

    def effect_nji(self, data=None):
        y_nji = self._prepare4est(data=data)

        if y_nji.ndim == 3:
            n, y_d, x_d = y_nji.shape
        elif y_nji.ndim == 2:
            n, y_d = y_nji.shape
            x_d = 1
        else:
            n, y_d, x_d = y_nji.shape[0], 1, 1

        y_nji = y_nji.reshape(n, y_d, x_d)
        zeros_ = np.zeros((n, y_d, 1))
        y_nji = np.concatenate((zeros_, y_nji), axis=2)

        return y_nji

    def _prepare4est(self, data=None):
        if not self._is_fitted:
            raise Exception("The estimator has not been fitted yet.")

        if data is None:
            wv = self._wv
        else:
            w, v = convert2array(data, self.adjustment, self.covariate)
            # if w is None:
            #     w = self._w
            wv = get_wv(w, v)

        if self.combined_treat:
            return self._prepare_combined_treat(wv)
        else:
            return self._prepare_separate_treat(wv)

    def _fit_combined_treat(self, x, wv, y, treat, control, categories, **kwargs):
        """Fit function which is used when multiple treatments are combined to
        give a single equivalent treatment vector.

        Parameters
        ----------
        x : np.array
            Treatment variables with shape (n, x_d)

        wv : np.array
            Covariate variables with shape (n, wv_d)

        y : np.array
            Outcome vevariablesctor with shape (n, y_d)

        treat : int or ndarray, optional
            If there is only one treament, then treat indicates the treatment
            group. If there are multiple treatment groups, then treat should
            be an ndarray of str with length equal to the number of treatments.
            For example, when there are multiple
            discrete treatments, array(['run', 'read']) means the treat value of
            the first treatment is taken as 'run' and that of the second treatment
            is taken as 'read'.

        control : int or list, optional
            See treat for more information

        categories : str

        Returns
        -------
        instance of TLearner
        """
        # specify the treat and control label
        x = self._comp_transformer(x, categories=categories)

        treat = get_tr_ctrl(
            treat,
            self._comp_transformer,
            treat=True,
            one_hot=False,
            discrete_treat=True,
        )
        control = get_tr_ctrl(
            control,
            self._comp_transformer,
            treat=False,
            one_hot=False,
            discrete_treat=True,
        )
        # treat = get_treat_control(treat, n_treatments, True)
        # control = get_treat_control(control, n_treatments, False)
        self.treat = treat
        self.control = control

        wv_treat, y_treat = get_groups(treat, x, True, wv, y)
        wv_control, y_control = get_groups(control, x, True, wv, y)

        y_treat = y_treat.squeeze()
        y_control = y_control.squeeze()

        self.xt_model.fit(wv_treat, y_treat, **kwargs)
        self.x0_model.fit(wv_control, y_control, **kwargs)

        self._is_fitted = True

        return self

    def _comp_transformer(self, x, categories="auto"):
        """Transform the discrete treatment into one-hot vectors.

        Parameters
        ----------
        x : ndarray, shape (n, x_d)
            An array containing the information of the treatment variables

        categories : str or list, optional
            by default 'auto

        Returns
        -------
        ndarray
            The transformed one-hot vectors
        """
        if x.shape[1] > 1:
            # if not self._is_fitted:
            if not hasattr(self, "ord_transformer"):
                self.ord_transformer = OrdinalEncoder(categories=categories)
                self.ord_transformer.fit(x)

                self.group_categories = self.ord_transformer.categories_

                labels = [np.arange(len(c)) for c in self.group_categories]
                labels = cartesian(labels)
                categories = [np.arange(len(labels))]

                self.label_dict = {tuple(k): i for i, k in enumerate(labels)}

            x_transformed = self.ord_transformer.transform(x).astype(int)
            x = np.full((x.shape[0], 1), np.NaN)

            for i, x_i in enumerate(x_transformed):
                x[i] = self.label_dict[tuple(x_i)]

        if not hasattr(self, "oh_transformer"):
            self.oh_transformer = OneHotEncoder()
            self.oh_transformer.fit(x)

        x = self.oh_transformer.transform(x).toarray()

        return x

    def _fit_separate_treat(self, x, wv, y, categories, **kwargs):
        """Fit function which is used when multiple treatments are treated separately,
        i.e., when combined_treat is False. For example, if there are 5 different
        discrete treatments, then we will fit 5 different models for them, respectivley.

        Parameters
        ----------
        x : ndarray
            Treatment variables.

        wv : ndarray
            Covariate variables.

        y : ndarray
            Outcome variables.

        group_categories : ndarray of ndarray
            Each ndarray in the group_categories is the classes of a single
            treatment vector. For example,
                array([array(['good', 'bad']), array([array(['a', 'b'])])])
            means there are two treatment vectors where the first treatment
            vector has classes 'good' and 'bad' while the second treatment
            vector has classes 'a' and 'b'.

        Returns
        -------
        instance of TLearner
        """
        self.transformer = OrdinalEncoder(categories=categories)
        self.transformer.fit(x)
        x = self.transformer.transform(x)

        self.group_categories = self.transformer.categories_

        self._fitted_dict_separa = defaultdict(list)
        waited_treat = [np.arange(len(i)) for i in self.group_categories]
        treat_arrays = cartesian(waited_treat)

        for treat in treat_arrays:
            model = clone(self.xt_model)
            _wv, _y = get_groups(treat, x, False, wv, y)
            _y = _y.squeeze()

            model.fit(_wv, _y, **kwargs)
            self._fitted_dict_separa["treatment"].append(treat)
            self._fitted_dict_separa["models"].append(model)

        self._is_fitted = True

        return self

    def _prepare_combined_treat(self, wv):
        yt = self.xt_model.predict(wv)
        y0 = self.x0_model.predict(wv)
        return yt - y0

    def _prepare_separate_treat(self, wv):
        n_treatments = len(self._fitted_dict_separa["treatment"])
        n = wv.shape[0]
        f_nji = np.full((n, self._y_d, n_treatments - 1), np.NaN)
        f_nj0 = self._fitted_dict_separa["models"][0].predict(wv)

        for i, model in enumerate(self._fitted_dict_separa["models"][1:]):
            fnji = (model.predict(wv) - f_nj0).reshape(n, self._y_d)
            f_nji[:, :, i] = fnji

        return f_nji.squeeze()

    # def __repr__(self) -> str:
    #     return f'TLearner'


class XLearner(BaseEstModel):
    """
    The XLearner is composed of 3 steps:
        1. Train two different models for the control group and treated group
            f_0(w), f_1(w)
        2. Generate two new datasets (h_0, w) using the control group and
            (h_1, w) using the treated group where
            h_0 = f_1(w) - y_0(w), h_1 = y_1(w) - f_0(w). Then train two models
            k_0(w) and k_1(w) in these datasets.
        3. Get the final model using the above two models:
            g(w) = k_0(w)a(w) + k_1(w)(1 - a(w)).
    Finally,  we estimate the ATE as follows:
        ATE = E_w(g(w)).
    See Kunzel, et al., (https://arxiv.org/abs/1706.03461) for reference.

    Attributes
    ----------
    ft_model : estiamtor, optional
        Machine learning model for the treatment gropu in the step 1.

    f0_model : estiamtor, optional
        Machine learning model for the control gropu in the step 1.

    kt_model : estiamtor, optional
        Machine learning model for the treatment gropu in the step 2.

    k0_model : estiamtor, optional
        Machine learning model for the control gropu in the step 2.

    _is_fitted : bool
        True if the instance of XLearner is fitted otherwise False.

    categories : str

    treatment : list of str, optional
        Names of the treatments.

    outcome : list of str, optional
        Names of the outcomes.

    adjustment : list of str, optional
        Names of the adjustment set.

    covariate : list of str, optional
        Names of the covariates.

    combined_treat : bool
        Whether use the combined treat technique when training the model.

    Methods
    ----------
    fit(data, outcome, treatment, adjustment, covariate, treat, control, combined_treatment)

    estimate(data, quantity=None)

    _fit_combined_treat(x, wv, y, treat, control, categories, **kwargs)
        Fit function when combined_treat is set to True.

    _comp_transformer(x, categories='auto')
        Transform the discrete treatment into one-hot vectors when combined_treat
        is set to True.

    _fit_separate_treat(x, wv, y, categories)
        Fit function when combined_treat is set to False.

    _prepare4est(data, outcome, treatment, adjustment, individual=None)
        Prepare (fit the model) for estimating various quantities including
        ATE, CATE, ITE, and CITE.

    _prepare_combined_treat(wv)

    _prepare_separate_treat(wv)
    """

    def __init__(
        self,
        model,
        random_state=2022,
        is_discrete_treatment=True,
        categories="auto",
        **kwargs,
    ):
        """
        Parameters
        ----------
        model : estimator, optional
            The base machine learning model for training XLearner. Any model
            should be some valid machine learning model with fit() and
            predict() functions.

        random_state : int

        categories : str
        """
        assert is_discrete_treatment is True

        self.model = clone(model)
        self.ft_model = clone(model)
        self.f0_model = clone(model)
        self.kt_model = clone(model)
        self.k0_model = clone(model)
        self._is_fitted = False

        super().__init__(
            random_state=random_state,
            is_discrete_treatment=is_discrete_treatment,
            categories=categories,
            **kwargs,
        )

    def fit(
        self,
        data,
        outcome,
        treatment,
        adjustment=None,
        covariate=None,
        treat=None,
        control=None,
        combined_treatment=True,
        **kwargs,
    ):
        """Fit the XLearner in the dataset.

        Parameters
        ----------
        data : pandas.DataFrame
            Training dataset for training the estimator.

        outcome : list of str, optional.
            List of names of the outcome features.

        treatment : list of str, optional
            List of names of the treatment features.

        adjustment : List of str, optional
            Lisf ot names of adjustment set ensuring the unconfoundness, by default None

        covariate : List of str, optional
            Covariate features, by default None

        treat : int, optional
            Label of the intended treatment group, by default None

        control : int, optional
            Label of the intended control group, by default None

        combined_treatment : bool, optional
            Only modify this parameter for multiple treatments, where multiple discrete
            treatments are combined to give a single new group of discrete treatment if
            set as True.
            When combined_treatment is set to True, then if there are multiple
            treatments, we can use the combined_treatment technique to covert
            the multiple discrete classification tasks into a single discrete
            classification task. For an example, if there are two different
            binary treatments:
                treatment_1: x_1 | x_1 \in {'sleep', 'run'},
                treatment_2: x_2 | x_2 \in {'study', 'work'},
            then we can convert these two binary classification tasks into
            a single classification task with 4 different classes:
                treatment: x | x \in {0, 1, 2, 3},
            where, for example, 1 stands for ('sleep' and 'stuy').

        Returns
        -------
        instance of XLearner
        """
        assert (
            adjustment is not None or covariate is not None
        ), "Need adjustment set or covariates to perform estimation."

        super().fit(
            data,
            outcome,
            treatment,
            adjustment=adjustment,
            covariate=covariate,
        )

        self.combined_treat = combined_treatment

        y, x, w, v = convert2array(data, outcome, treatment, adjustment, covariate)
        self._y_d = y.shape[1]

        if self.categories == "auto" or self.categories is None:
            categories = "auto"
        else:
            categories = list(self.categories)

        # self.transformer = OrdinalEncoder(categories=categories)
        # self.transformer.fit(x)
        # x = self.transformer.transform(x)

        # self.group_categories = self.transformer.categories_
        wv = get_wv(w, v)
        self._wv = wv
        self._w = w

        if combined_treatment:
            return self._fit_combined_treat(
                x, wv, y, treat, control, categories, **kwargs
            )
        else:
            return self._fit_separate_treat(x, wv, y, categories, **kwargs)

    def _prepare4est(self, data=None, rho=0.5, *args, **kwargs):
        if not self._is_fitted:
            raise Exception("The estimator has not been fitted yet.")

        if data is None:
            wv = self._wv
        else:
            w, v = convert2array(data, self.adjustment, self.covariate)
            # if w is None:
            #     w = self._w
            wv = get_wv(w, v)

        if self.combined_treat:
            effect = self._prepare_combined_treat(wv, rho)
        else:
            effect = self._prepare_separate_treat(wv, rho)

        return effect

    def estimate(self, data=None, rho=0.5, quantity=None, *args, **kwargs):
        # TODO: add support for other types of rho
        effect = self._prepare4est(data, rho, *args, **kwargs)

        if quantity == "CATE" or quantity == "ATE":
            return np.mean(effect, axis=0)
        else:
            return effect

    def effect_nji(self, data=None):
        y_nji = self._prepare4est(data=data)

        if y_nji.ndim == 3:
            n, y_d, x_d = y_nji.shape
        elif y_nji.ndim == 2:
            n, y_d = y_nji.shape
            x_d = 1
        else:
            n, y_d, x_d = y_nji.shape[0], 1, 1

        y_nji = y_nji.reshape(n, y_d, x_d)
        zeros_ = np.zeros((n, y_d, 1))
        y_nji = np.concatenate((zeros_, y_nji), axis=2)

        return y_nji

    def _fit_combined_treat(self, x, wv, y, treat, control, categories, **kwargs):
        """Fit function which is used when multiple treatments are combined to
        give a single equivalent treatment vector.

        Parameters
        ----------
        x : np.array
            Treatment variables with shape (n, x_d)

        wv : np.array
            Covariate variables with shape (n, wv_d)

        y : np.array
            Outcome vevariablesctor with shape (n, y_d)

        treat : int or list, optional
            If there is only one treament, then treat indicates the treatment
            group. If there are multiple treatment groups, then treat should
            be a list of str with length equal to the number of treatments.
            For example, when there are multiple
            discrete treatments, array(['run', 'read']) means the treat value of
            the first treatment is taken as 'run' and that of the second treatment
            is taken as 'read'.

        control : int or list, optional
            See treat for more information

        categories : str

        Returns
        -------
        instance of XfLearner
        """
        x = self._comp_transformer(x, categories)

        treat = get_tr_ctrl(
            treat,
            self._comp_transformer,
            treat=True,
            one_hot=False,
            discrete_treat=True,
        )
        control = get_tr_ctrl(
            control,
            self._comp_transformer,
            treat=False,
            one_hot=False,
            discrete_treat=True,
        )
        self.treat = treat
        self.control = control

        # Step 1
        # TODO: modify the generation of group ids for efficiency
        wv_treat, y_treat = get_groups(treat, x, True, wv, y)
        wv_control, y_control = get_groups(control, x, True, wv, y)

        y_treat = y_treat.squeeze()
        y_control = y_control.squeeze()

        self.ft_model.fit(wv_treat, y_treat, **kwargs)
        self.f0_model.fit(wv_control, y_control, **kwargs)

        # Step 2
        h_treat_target = y_treat - self.f0_model.predict(wv_treat)
        h_control_target = self.ft_model.predict(wv_control) - y_control
        self.kt_model.fit(wv_treat, h_treat_target)
        self.k0_model.fit(wv_control, h_control_target)

        # Step 3
        # This is the task of predict. See _prepare4est

        self._is_fitted = True

        return self

    def _comp_transformer(self, x, categories="auto"):
        """Transform the discrete treatment into one-hot vectors.

        Parameters
        ----------
        x : ndarray, shape (n, x_d)
            An array containing the information of the treatment variables

        categories : str or list, optional
            by default 'auto'

        Returns
        -------
        ndarray
            The transformed one-hot vectors
        """
        if x.shape[1] > 1:
            # if not self._is_fitted:
            if not hasattr(self, "ord_transformer"):
                self.ord_transformer = OrdinalEncoder(categories=categories)
                self.ord_transformer.fit(x)

                self.group_categories = self.ord_transformer.categories_

                labels = [np.arange(len(c)) for c in self.group_categories]
                labels = cartesian(labels)
                categories = [np.arange(len(labels))]

                self.label_dict = {tuple(k): i for i, k in enumerate(labels)}

            x_transformed = self.ord_transformer.transform(x).astype(int)
            x = np.full((x.shape[0], 1), np.NaN)

            for i, x_i in enumerate(x_transformed):
                x[i] = self.label_dict[tuple(x_i)]

        if not hasattr(self, "oh_transformer"):
            self.oh_transformer = OneHotEncoder()
            self.oh_transformer.fit(x)

        x = self.oh_transformer.transform(x).toarray()

        return x

    def _fit_separate_treat(self, x, wv, y, categories, *args, **kwargs):
        """Fit function which is used when multiple treatments are treated separately,
        i.e., when combined_treat is False.

        Parameters
        ----------
        x : ndarray
            Treatment variables.

        wv : ndarray
            Covariate variables.

        y : ndarray
            Outcome variables.

        group_categories : ndarray of ndarray
            Each ndarray in the group_categories is the classes of a single
            treatment vector. For example,
                array([array(['good', 'bad']), array([array(['a', 'b'])])])
            means there are two treatment vectors where the first treatment
            vector has classes 'good' and 'bad' while the second treatment
            vector has classes 'a' and 'b'.

        Returns
        -------
        instance of XLearner
        """
        self.transformer = OrdinalEncoder(categories=categories)
        self.transformer.fit(x)
        x = self.transformer.transform(x)

        self.group_categories = self.transformer.categories_

        self._fitted_dict_separa = defaultdict(list)
        waited_treat = [np.arange(len(i)) for i in self.group_categories]
        treat_arrays = cartesian(waited_treat)

        f0_model = clone(self.f0_model)
        _wv_control, _y_control = get_groups(treat_arrays[0], x, False, wv, y)
        _y_control = _y_control.squeeze()
        f0_model.fit(_wv_control, _y_control)

        for treat in treat_arrays[1:]:
            ft_model = clone(self.ft_model)
            _wv, _y = get_groups(treat, x, False, wv, y)
            _y = _y.squeeze()

            # Step 1
            ft_model.fit(_wv, _y, *args, **kwargs)

            # Step 2
            h_treat_target = _y - f0_model.predict(_wv)
            h_control_target = ft_model.predict(_wv_control) - _y_control
            kt_model = clone(self.kt_model)
            k0_model = clone(self.k0_model)
            kt_model.fit(_wv, h_treat_target)
            k0_model.fit(_wv_control, h_control_target)

            # Step 3
            self._fitted_dict_separa["models"].append((kt_model, k0_model))
            self._fitted_dict_separa["treatment"].append(treat)

        self._is_fitted = True

        return self

    def _prepare_combined_treat(self, wv, rho):
        # TODO: add support for training to select rho
        kt_pred = self.kt_model.predict(wv)
        k0_pred = self.k0_model.predict(wv)

        return rho * kt_pred + (1 - rho) * k0_pred

    def _prepare_separate_treat(self, wv, rho):
        model_list = self._fitted_dict_separa["models"]
        n_treatments = len(self._fitted_dict_separa["treatment"])
        n = wv.shape[0]
        f_nji = np.full((n, self._y_d, n_treatments), np.NaN)

        for i, (kt_model, k0_model) in enumerate(model_list):
            pred_t = kt_model.predict(wv)
            pred_0 = k0_model.predict(wv)
            fnji = (rho * pred_t + (1 - rho) * pred_0).reshape(n, self._y_d)
            f_nji[:, :, i] = fnji

        return f_nji.squeeze()

    # def __repr__(self) -> str:
    #     return f'XLearner'



try:
    from packaging.version import Version
except ModuleNotFoundError:
    from distutils.version import LooseVersion as Version

from ._common import const, set_random_state, infer_task_type, unique, to_repr
from ._common import to_df, context, is_notebook, view_pydot
from ._common import to_snake_case, to_camel_case, drop_none



from copy import deepcopy
from collections import defaultdict
from turtle import clone

import networkx as nx
import numpy as np

from sklearn.linear_model import LinearRegression as LR

from ylearn.utils._common import check_cols
from ylearn.estimator_model.double_ml import DML4CATE
from .utils import (check_nodes, check_ancestors_chain,
                    powerset, IdentificationError, remove_ingo_edges,
                    descendents_of_iter)
from .prob import Prob

np.random.seed(2022)


class CausalModel:
    """
    Basic object for performing causal inference.

    Attributes
    ----------
    data : pandas.DataFrame
        Data used for discovering causal graph and training estimator model.
    
    causal_graph : CausalGraph #TODO: support CausalStructuralModel
    
    treatment : list of str, optional
        Names of treatments in the current identification problem.
    
    outcome : list of str, optional
        Names of outcomes in the current identification problem.
    
    ava_nodes : dictkeys
        All observed nodes of the CausalGraph
    
    cached_result : dict
        Results of previous identification results.
    
    _adjustment_set : list of set
        All possible backdoor adjustment set

    Methods
    ----------
    id(y, x, prob=None, graph=None)
        Identify the causal quantity P(y|do(x)) in the graph if identifiable
        else return False, where y can be a set of different outcomes and x
        can be a set of different treatments. #TODO: be careful about that
        currently we treat a random variable equally as its value, i.e., we
        do not discern P(Y|do(X)) and P(Y=y|do(X=x)).
    
    identify(treatment, outcome, identify_method='auto')
        Identify the causal effect of treatment on outocme expression.
    
    estimate(estimator_model, data, adjustment, covariate, quantity=None)
        Estimate the causal effect of treatment on outcome (y).
    
    get_iv(treatment, outcome)
        Find the instrumental variables for the causal effect of the
        treatment on the outcome.
    
    is_iv(treatment, outcome, set_)
        Determine whether a given set_ is a valid instrumental variable set.
    
    discover_graph(data)
        Perform causal discovery over data.
    
    is_valid_backdoor_set(set, treatment, outcome)
        Determine if a given set is a valid backdoor adjustment set for causal
        effects of treatments on outcomes.
    
    get_backdoor_set(treatment, outcome, adjust='simple')
        Return backdoor adjustment sets for the given treatment and outcome
        in the style stored in adjust (simple, minimal, or all).
    
    get_backdoor_path(treatment, outcome)
        Return all backdoor paths in the graph between treatment and outcome.
    
    has_collider(path, backdoor_path=True)
        If the path in the current graph has a collider, return True, else
        return False.
    
    is_connected_backdoor_path(path)
        Test whether a backdoor path is connected in the current graph, where
        path[0] is the start of the path and path[-1] is the end.
    
    is_frontdoor_set(set, treatment, outcome)
        Determine if a given set is a frontdoor adjustment set.
    
    get_frontdoor_set(treatment, outcome, adjust)
        Return the frontdoor adjustment set for given treatment and outcome.
    
    estimate_hidden_cofounder()
        Estimation of hidden cofounders.
    """

    def __init__(self, causal_graph=None, data=None, **kwargs):
        """
        Parameters
        ----------
        causal_graph : CausalGraph
            An instance of CausalGraph which encodes the causal structures.
        
        data : DataFrame (for now)
        """
        self.data = data
        self._adjustment_set = None
        self.treatment = None
        self.outcome = None
        self.cached_result = defaultdict(list)

        self.causal_graph = causal_graph if causal_graph is not None\
            else self.discover_graph(data)

        # self.ava_nodes = self.causal_graph.causation.keys()

    def id(self, y, x, prob=None, graph=None):
        """Identify the causal quantity P(y|do(x)) if identifiable else return
        False. See Shpitser and Pearl (2006b)
        (https://ftp.cs.ucla.edu/pub/stat_ser/r327.pdf) for reference.
        Note that here we only consider semi-Markovian causal model, where
        each unobserved variable is a parent of exactly two nodes. This is
        because any causal model with unobserved variables can be converted
        to a semi-Markovian causal model encoding the same set of conditional
        independences (Verma, 1993).

        Parameters
        ----------
        y : set of str
            Set of names of outcomes.
        
        x : set of str
            Set of names of treatments.
        
        prob : Prob
            Probability distribution encoded in the graph.
        
        graph : CausalGraph

        Returns
        ----------
        Prob if identifiable.

        Raises
        ----------
        IdentificationError if not identifiable.
        """
        # TODO: need to be careful about the fact that sets are not ordered,
        # be careful about the usage of list and set in the current
        # implementation
        if graph is None:
            graph = self.causal_graph

        if prob is None:
            prob = graph.prob

        v = set(graph.causation.keys())
        v_topo = list(graph.topo_order)

        y, x = set(y), set(x)

        # 1
        if not x:
            if prob.divisor or prob.product:
                prob.marginal = v.difference(y).union(prob.marginal)
            else:
                # If the prob is not built with products, then
                # simply replace the variables with y
                prob.variables = y
            return prob

        # 2
        ancestor = graph.ancestors(y)
        prob_ = deepcopy(prob)
        if v.difference(ancestor) != set():
            an_graph = graph.build_sub_graph(ancestor)
            if prob_.divisor or prob_.product:
                prob_.marginal = v.difference(ancestor).union(prob_.marginal)
            else:
                prob_.variables = ancestor
            return self.id(y, x.intersection(ancestor), prob_, an_graph)

        # 3
        w = v.difference(x).difference(
            graph.remove_incoming_edges(x, new=True).ancestors(y)
        )
        if w:
            return self.id(y, x.union(w), prob, graph)

        # 4
        c = list(graph.remove_nodes(x, new=True).c_components)
        if len(c) > 1:
            product_expressioin = set()
            for subset in c:
                product_expressioin.add(
                    self.id(subset, v.difference(subset), prob, graph)
                )
            return Prob(
                marginal=v.difference(y.union(x)), product=product_expressioin
            )
        else:
            s = c.pop()
            cg = list(graph.c_components)
            # 5
            if cg[0] == set(graph.dag.nodes):
                raise IdentificationError(
                    'The causal effect is not identifiable in the'
                    'current graph.'
                )
            # 6
            elif s in cg:
                product_expression = set()
                for element in s:
                    product_expression.add(
                        Prob(variables={element},
                             conditional=set(v_topo[:v_topo.index(element)]))
                    )
                return Prob(
                    marginal=s.difference(y), product=product_expression
                )

            # 7
            else:
                # TODO: not clear whether directly replacing a random variable
                # with one of its value matters in this line
                for subset in cg:
                    if s.intersection(subset) == s:
                        product_expressioin = set()
                        for element in subset:
                            product_expressioin.add(
                                Prob(variables={element},
                                     conditional=set(v_topo[
                                         :v_topo.index(element)
                                     ]))
                            )
                        sub_prob = Prob(product=product_expressioin)
                        sub_graph = graph.build_sub_graph(subset)
                        return self.id(
                            y, x.intersection(subset), sub_prob, sub_graph
                        )

    def identify(self, treatment, outcome, identify_method='auto'):
        """Identify the causal effect expression. Identification is an operation that
        converts any causal effect quantity, e.g., quantities with the do operator, into
        the corresponding statistical quantity such that it is then possible
        to estimate the causal effect in some given data. However, note that not all
        causal quantities are identifiable, in which case an IdentificationError
        will be raised.

        Parameters
        ----------
        treatment : set or list of str, optional
            Set of names of treatments.
        
        outcome : set or list of str, optional
            Set of names of outcomes.
        
        identify_method : tuple of str or str, optional. Default to 'auto'.
            If the passed value is a tuple or list, then it should have two
            elements where the first one is for the identification methods
            and the second is for the returned set style.

            Available options:
            'auto' : Perform identification with all possible methods
            'general': The general identification method, see id()
            ('backdoor', 'simple'): Return the set of all direct confounders of
                                    both treatments and outcomes as a backdoor
                                    adjustment set.
            ('backdoor', 'minimal'): Return all possible backdoor adjustment sets with
                                     minial number of elements.
            ('backdoor', 'all'): Return all possible backdoor adjustment sets.
            ('frontdoor', 'simple'): Return all possible frontdoor adjustment sets with
                                      minial number of elements.
            ('frontdoor', 'minimal'): Return all possible frontdoor adjustment sets with
                                      minial number of elements.
            ('frontdoor', 'all'): Return all possible frontdoor adjustment sets.

        Returns
        ----------
        dict
            Keys of the dict are identify methods while the values are the
            corresponding results.

        # TODO: support finding the minimal general adjustment set in linear
        # time
        """
        assert treatment and outcome, 'Please provide names of treatments and outcomes.'

        # check nodes avaliable in the graph
        ava_nodes = self.causal_graph.causation.keys()
        check_nodes(ava_nodes, treatment, outcome)

        self.treatment = treatment
        self.outcome = outcome

        if identify_method == 'auto':
            result_dict = {
                'ID': self.id(outcome, treatment),
                'backdoor': self.get_backdoor_set(treatment, outcome, 'simple'),
                'frontdoor': self.get_frontdoor_set(treatment, outcome, 'simple')
            }
            self.cached_result[f'Treatment: {treatment}, Outcome: {outcome}'].append(
                result_dict
            )
            return result_dict

        if identify_method == 'general':
            result = self.id(outcome, treatment)
            self.cached_result[f'Treatment: {treatment}, Outcome: {outcome}'].append(
                {identify_method: result}
            )
            return {'ID': result}

        if identify_method[0] == 'backdoor':
            adjustment = self.get_backdoor_set(
                treatment, outcome, adjust=identify_method[1]
            )
            self.cached_result[f'Treatment: {treatment}, Outcome: {outcome}'].append(
                {identify_method[0]: adjustment}
            )
        elif identify_method[0] == 'frontdoor':
            adjustment = self.get_frontdoor_set(
                treatment, outcome, adjust=identify_method[1]
            )
            self.cached_result[f'Treatment: {treatment}, Outcome: {outcome}'].append(
                {identify_method[0]: adjustment}
            )
        else:
            raise IdentificationError(
                'Support backdoor adjustment, frontdoor adjustment, and general'
                f'identification, but was given {identify_method[1]}'
            )

        return {f'{identify_method[0]}': adjustment}

    def estimate(
        self,
        estimator_model,
        data=None, *,
        treatment=None,
        outcome=None,
        adjustment=None,
        covariate=None,
        quantity=None,
        **kwargs
    ):
        """Estimate the identified causal effect in a new dataset.

        Parameters
        ----------
        estimator_model : EstimatorModel
            Any suitable estimator models implemented in the EstimatorModel can
            be applied here. 
        
        data : pandas.Dataframe, optional. Default to None
            The data set for causal effect to be estimated. If None, use the data
            which is used for discovering causal graph.
        
        treatment : set or list, optional. Default to None
            Names of the treatment. If None, the treatment used for backdoor adjustment
            will be taken as the treatment.
            
        outcome : set or list, optional. Default to None
            Names of the outcome. If None, the treatment used for backdoor adjustment
            will be taken as the outcome.
        
        adjustment : set or list, optional. Default to None
            Names of the adjustment set. If None, the ajustment set is given by
            the simplest backdoor set found by CausalModel.
        
        covariate : set or list, optional. Default to None
            Names of covariate set. Ignored if set as None.
        
        quantity : str, optional. Default to None
            The interested quantity when evaluating causal effects.

        Returns
        -------
        ndarray or float
            The estimated causal effect in data.
        """
        data = self.data if data is None else data
        check_cols(data, treatment, outcome, adjustment, covariate)

        ava_nodes = self.causal_graph.causation.keys()
        check_nodes(ava_nodes, adjustment, covariate)

        # assert self.outcome is not None
        # assert self.treatment is not None

        outcome = self.outcome if outcome is None else outcome
        treatment = self.treatment if treatment is None else treatment

        # make sure the adjustment and the covariate are valid backdoor adjustment sets
        if adjustment is None and covariate is None:
            if self._adjustment_set is None:
                adjustment = self.get_backdoor_set(
                    self.treatment, self.outcome, 'simple'
                )[0]
            else:
                adjustment = self._adjustment_set[0]
        else:
            if adjustment is not None and covariate is not None:
                temp_set = set(adjustment).union(set(covariate))
            else:
                temp_set = [
                    set(x) for x in [adjustment, covariate] if x is not None
                ]
                temp_set = temp_set.pop()

            assert self.is_valid_backdoor_set(
                temp_set, self.treatment, self.outcome
            ), 'The adjustment set should be a valid backdoor adjustment set,'
            f'but was given {temp_set}.'

        # fit the estimator_model if it is not fitted
        if not estimator_model._is_fitted:
            estimator_model.fit(
                data=data,
                outcome=outcome,
                treatment=treatment,
                covariate=covariate,
                adjustment=adjustment,
                **kwargs,
            )

        effect = estimator_model.estimate(data=data, quantity=quantity)

        return effect

    def identify_estimate(
        self,
        data,
        outcome,
        treatment,
        estimator_model=None,
        quantity=None,
        identify_method='auto',
        **kwargs
    ):
        """Combination of the identifiy method and the estimate method. However,
        since current implemented estimator models assume (conditionally)
        unconfoundness automatically (except for methods related to iv), we may
        only consider using backdoor set adjustment to fullfill the unconfoundness
        condition.

        Parameters
        ----------
        data : pandas.DataFrame
            The data used for training the estimator models and for estimating the
            causal effect.
        
        treatment : set or list of str, optional
            Set of names of treatments.
        
        outcome : set or list of str, optional
            Set of names of outcomes.
        
        identify_method : tuple of str or str, optional. Default to 'auto'.
            If the passed value is a tuple or list, then it should have two
            elements where the first one is 'backdoor' and the second is for
            the returned set style.

            Available options:
            'auto' : Find all possible backdoor adjustment set.
            ('backdoor', 'simple') or 'simple': Return the set of all direct confounders of
                                                both treatments and outcomes as a backdoor
                                                adjustment set.
            ('backdoor', 'minimal') or 'minimal': Return all possible backdoor adjustment sets with
                                                    minial number of elements.
            ('backdoor', 'all') or 'all': Return all possible backdoor adjustment sets.
        
        quantity : str, optional. Default to None
            The interested quantity when evaluating causal effects.

        Returns
        ----------
        ndarray or float
            The estimated causal effect in data.
        """
        # TODO: now only supports estimation with adjustment set. This needs
        # to be updated if the estimation of general identification problem
        # is solved.
        assert all((data is not None, treatment is not None, outcome is not None))

        # if estimator_model is None:
        #     estimator_model = DML4CATE()
        assert estimator_model is not None

        # check columns
        check_cols(data, treatment, outcome)

        # check nodes in the graph
        ava_nodes = self.causal_graph.causation.keys()
        check_nodes(ava_nodes, treatment, outcome)

        if isinstance(identify_method, (list, tuple)):
            identify_method = identify_method[1]

        adj_set = self.get_backdoor_set(
            treatment, outcome, adjust=identify_method
        )

        self._est_result_dict = defaultdict(list)

        for sub_adj_set in adj_set:
            est_model = clone(estimator_model)

            est_model.fit(
                data=data,
                outcome=outcome,
                treatment=treatment,
                covariate=sub_adj_set,
                **kwargs,
            )

            self._est_result_dict['est_models'].append(est_model)

            effect = estimator_model.estimate(data=data, quantity=quantity)
            self._est_result_dict['effects'].append(effect)

        return self._est_result_dict

    def discover_graph(self, data):
        """Discover the causal graph from data.

        Parameters
        ----------
        data : pandas.DataFrame

        Returns
        ----------
        CausalGraph
        """
        assert data is not None, 'Need data to call causal discovery.'
        raise NotImplementedError

    def is_valid_backdoor_set(self, set_, treatment, outcome):
        """Determine if a given set is a valid backdoor adjustment set for
        causal effect of treatments on the outcomes.

        Parameters
        ----------
        set_ : set
            The adjustment set.
        
        treatment : set or list of str
            str is also acceptable for single treatment.
        
        outcome : set or list of str
            str is also acceptable for single outcome.

        Returns
        ----------
        bool
            True if the given set is a valid backdoor adjustment set for the 
            causal effect of treatment on outcome in the current causal graph.
        """
        # TODO: improve the implementation
        # A valid backdoor set d-separates all backdoor paths between any pairs
        # of treatments and outcomes.
        treatment = set(treatment) if not isinstance(treatment, str) \
            else {treatment}
        outcome = set(outcome) if not isinstance(outcome, str) else {outcome}

        # make sure all nodes are in the graph
        ava_nodes = self.causal_graph.causation.keys()
        check_nodes(ava_nodes, treatment, outcome, set_)

        # build the causal graph with unobserved variables explicit
        modified_dag = self.causal_graph.remove_outgoing_edges(
            treatment, new=True
        ).explicit_unob_var_dag

        return nx.d_separated(modified_dag, treatment, outcome, set_)

    def get_backdoor_set(
        self,
        treatment,
        outcome,
        adjust='simple',
        print_info=False
    ):
        """Return the backdoor adjustment set for the given treatment and outcome.

        Parameters
        ----------
        treatment : set or list of str
            Names of the treatment. str is also acceptable for single treatment.
        
        outcome : set or list of str 
            Names of the outcome. str is also acceptable for single outcome.
        
        adjust : str
            Set style of the backdoor set
                simple: directly return the parent set of treatment
                minimal: return the minimal backdoor adjustment set
                all: return all valid backdoor adjustment set.
        print_info : bool
            If True, print the identified results.

        Raises
        ----------
        Exception : IdentificationError
            Raise error if the style is not in simple, minimal or all or no
            set can satisfy the backdoor criterion.

        Returns
        ----------
        tuple
            The first element is the adjustment list, while the second is the
            encoded Prob.
        """
        # TODO: can I find the adjustment sets by using the adj matrix
        # TODO: improve the implementation

        # convert treatment and outcome to sets.
        treatment = set(treatment) if type(treatment) is not str \
            else {treatment}
        outcome = set(outcome) if type(outcome) is not str else {outcome}

        # make sure all nodes are present
        ava_nodes = self.causal_graph.causation.keys()
        check_nodes(ava_nodes, outcome, treatment)

        modified_dag = self.causal_graph.remove_outgoing_edges(
            treatment, new=True
        ).explicit_unob_var_dag

        def determine(modified_dag, treatment, outcome):
            if all(
                [self.causal_graph.causation[t] == [] for t in treatment]
            ):
                return nx.d_separated(modified_dag, treatment,
                                      outcome, set())
            return True

        # raise exception if no set can be used as the backdoor set
        if not determine(modified_dag, treatment, outcome):
            print(
                'No set can satisfy the backdoor criterion.'
            )

            return None

        # Simply take all parents of both treatment and outcome as the backdoor set
        if adjust == 'simple':
            # TODO: if the parent of x_i is the descendent of another x_j
            backdoor_list = []
            for t in treatment:
                backdoor_list += self.causal_graph.causation[t]
            adset = set(backdoor_list)
        # use other methods
        else:
            # Get all backdoor sets. currently implenmented
            # in a brutal force manner. NEED IMPROVEMENT
            des_set, backdoor_set_list = set(), []
            for t in treatment:
                des_set.update(
                    nx.descendants(self.causal_graph.observed_dag, t)
                )
            initial_set = (
                set(list(ava_nodes)) -
                treatment - outcome - des_set
            )

            if adjust == 'minimal':
                # return the backdoor set which has the fewest number of elements
                for i in powerset(initial_set):
                    if nx.d_separated(modified_dag, treatment, outcome, i):
                        backdoor_set_list.append(i)
                        break
            else:
                backdoor_set_list = [
                    i for i in powerset(initial_set)
                    if nx.d_separated(modified_dag, treatment, outcome, i)
                ]

            if not backdoor_set_list and not nx.d_separated(
                modified_dag, treatment, outcome, set()
            ):
                raise IdentificationError(
                    'No set can satisfy the backdoor criterion.'
                )

            try:
                backdoor_set_list[0]
            except Exception:
                backdoor_set_list = [[]]
            finally:
                adset = set(backdoor_set_list[0])

            if any((adjust == 'all', adjust == 'minimal')):
                backdoor_list = backdoor_set_list
                self._adjustment_set = backdoor_set_list
            else:
                raise IdentificationError(
                    'Do not support backdoor set styles other than simple, all'
                    'or minimal.'
                )

        # Build the corresponding probability distribution.
        product_expression = {
            Prob(variables=outcome,
                 conditional=treatment.union(adset)),
            Prob(variables=adset)
        }
        prob = Prob(marginal=adset, product=product_expression)

        if print_info:
            print(
                f'The corresponding statistical estimand should be {prob.parse()})'
            )
        return (backdoor_list, prob)

    def get_backdoor_path(self, treatment, outcome):
        """Return all backdoor paths connecting treatment and outcome.

        Parameters
        ----------
        treatment : str
        
        outcome : str

        Returns
        ----------
        list
            A list containing all valid backdoor paths between the treatment and
            outcome in the graph.
        """
        dag = self.causal_graph.explicit_unob_var_dag

        ava_nodes = self.causal_graph.causation.keys()
        check_nodes(ava_nodes, treatment, outcome)

        return [
            p for p in nx.all_simple_paths(
                dag.to_undirected(), treatment, outcome
            )
            if len(p) > 2 and p[1] in dag.predecessors(treatment)
        ]

    def has_collider(self, path, backdoor_path=True):
        """If the path in the current graph has a collider, return True, else
        return False.

        Parameters
        ----------
        path : list of str
            A list containing nodes in the path.
        
        backdoor_path : bool
            Whether the path is a backdoor path

        Returns
        ----------
        Boolean
            True if the path has a collider.
        """
        # TODO: improve the implementation.
        dag = self.causal_graph.explicit_unob_var_dag
        check_nodes(dag.nodes, path)

        if path[1] not in dag.predecessors(path[0]):
            backdoor_path = False

        if len(path) > 2:
            assert(nx.is_path(dag.to_undirected(), path)), "Not a valid path."

            j = 0
            if backdoor_path:
                if len(path) == 3:
                    return False

                for i, node in enumerate(path[1:], 1):
                    if node in dag.successors(path[i-1]):
                        # if path[i] is the last element, then it is impossible
                        # to have collider
                        if i == len(path) - 1:
                            return False
                        j = i
                        break
                for i, node in enumerate(path[j+1:], j+1):
                    if node in dag.predecessors(path[i-1]):
                        return True
            else:
                for i, node in enumerate(path[1:], 1):
                    if node in dag.predecessors(path[i-1]):
                        return True

        return False

    def is_connected_backdoor_path(self, path):
        """Test whether a backdoor path is connected.

        Parameters
        ----------
        path : list
            A list describing the path.

        Returns
        ----------
        Boolean
            True if path is a d-connected backdoor path and False otherwise.
        """
        check_node_graph = self.causal_graph.explicit_unob_var_dag
        check_nodes(check_node_graph.nodes, path)

        assert path[1] in check_node_graph.predecessors(path[0]), \
            'Not a backdoor path.'

        # A backdoor path is not connected if it contains a collider or not a
        # backdoor_path.
        # TODO: improve the implementation
        if self.has_collider(path) or \
                path not in self.get_backdoor_path(path[0], path[-1]):
            return False
        return True

    def is_frontdoor_set(self, set_, treatment, outcome):
        """Determine if the given set is a valid frontdoor adjustment set for the
        causal effect of treatment on outcome.

        Parameters
        ----------
        set_ : set of str
        
        treatement : str
        
        outcome : str

        Returns
        ----------
        Bool
            True if the given set is a valid frontdoor adjustment set for causal effects
            of treatments on outcomes.
        """
        ava_nodes = self.causal_graph.causation.keys()
        check_nodes(ava_nodes, set_, set(treatment), set(outcome))

        # rule 1, intercept all directed paths from treatment to outcome
        for path in nx.all_simple_paths(
            self.causal_graph.observed_dag, treatment, outcome
        ):
            if not any([path_node in set_ for path_node in path]):
                return False

        # rule 2, there is no unblocked back-door path from treatment to set_
        for subset in set_:
            for path in self.get_backdoor_path(treatment, subset):
                if self.is_connected_backdoor_path(path):
                    return False

        # rule 3, all backdoor paths from set_ to outcome are blocked by
        # treatment
        return self.is_valid_backdoor_set({treatment}, set_, {outcome})

    def get_frontdoor_set(self, treatment, outcome, adjust='simple'):
        """Return the frontdoor set for adjusting the causal effect between
        treatment and outcome.

        Parameters
        ----------
        treatment : set of str or str
            Name of the treatment. Should contain only one element.
        
        outcome : set of str or str
            Name of the outcome. Should contain only one element.

        Returns
        ----------
        tuple
            2 elements (adjustment_set, Prob)
        """
        ava_nodes = self.causal_graph.causation.keys()
        check_nodes(ava_nodes, set(treatment), set(outcome))

        if type(treatment) is set and type(outcome) is set:
            assert (
                len(treatment) == 1 and len(outcome) == 1
            ), 'Treatment and outcome should be sets which contain one'
            'element for frontdoor adjustment.'
            treatment, outcome = treatment.pop(), outcome.pop()

        # Find the initial set which will then be used to generate all
        # possible frontdoor sets with the method powerset()
        initial_set = set()
        for path in nx.all_simple_paths(
            self.causal_graph.observed_dag, treatment, outcome
        ):
            initial_set.update(set(path))
        initial_set = initial_set - {treatment} - {outcome}
        potential_set = powerset(initial_set)

        # Different frontdoor set styles.
        if adjust == 'simple' or adjust == 'minimal':
            adjustment, adset = set(), set()
            for set_ in potential_set:
                if self.is_frontdoor_set(set_, treatment, outcome):
                    adjustment.update(set_)
                    adset.update(set_)
                    break
        elif adjust == 'all':
            adjustment = [
                i for i in powerset(initial_set) if self.is_frontdoor_set(
                    i, treatment, outcome
                )
            ]
            adset = adjustment[0] if adjustment else set()
        else:
            raise IdentificationError(
                'The frontdoor set style must be one of simple, all'
                'and minimal.'
            )

        if not adjustment:
            raise IdentificationError(
                'No set can satisfy the frontdoor adjustment criterion.'
            )

        # Build the corresponding probability distribution.
        product_expression = {
            Prob(variables=adset, conditional={treatment}),
            Prob(
                marginal={treatment}, product={
                    Prob(variables={outcome},
                         conditional={treatment}.union(adset)),
                    Prob(variables={treatment})
                }
            )
        }
        prob = Prob(marginal=adset, product=product_expression)
        return (adjustment, prob)

    def get_iv(self, treatment, outcome):
        """Find the instrumental variables for the causal effect of the
        treatment on the outcome.

        Parameters
        ----------
        treatment : iterable
            Name(s) of the treatment.
        
        outcome : iterable
            Name(s) of the outcome.

        Returns
        -------
        set
            A valid instrumental variable set which will be an empty one if
            there is no such set.
        """
        ava_nodes = self.causal_graph.causation.keys()
        check_nodes(ava_nodes, treatment, outcome)

        # 1. relevance: all possible instrument variables should be parents of the
        # treatment
        waited_instrument = set(self.causal_graph.parents(treatment))

        # 2. exclusion: build the graph where all incoming edges to the treatment
        # are removed such that those nodes which have effects on the outcome through
        # the treatment in the modified graph are excluded to be valid instruments
        exp_unob_graph = self.causal_graph.explicit_unob_var_dag
        modified_graph = remove_ingo_edges(exp_unob_graph, True, treatment)
        excluded_nodes_an = nx.ancestors(modified_graph, outcome)
        waited_instrument.difference_update(excluded_nodes_an)

        # We also should not count on descendents of ancestors of the outcome which
        # may have effects on y through backdoor paths
        # excluded_nodes_des = nx.descendants(modified_graph, excluded_nodes_an)
        excluded_nodes_des = descendents_of_iter(
            modified_graph, excluded_nodes_an
        )
        iv = waited_instrument.difference(excluded_nodes_des)

        if not iv:
            # print(f'No valid instrument variable has been found.')
            iv = None

        return iv

    def is_valid_iv(self, treatment, outcome, set_):
        """Determine whether a given set_ is a valid instrumental variable set.

        Parameters
        ----------
        treatment : iterable
            Name(s) of the treatment.
        
        outcome : iterable
            Name(s) of the outcome.
        
        set_ : iterable

        Returns
        -------
        bool
            True if the set is a valid instrumental variable set and False
            otherwise.
        """
        all_iv = self.get_iv(treatment, outcome)

        if not all_iv:
            raise Exception(
                'No valid instrument variable is found in the current causal model.'
            )

        set_ = {set_} if isinstance(set_, str) else set_

        for element in set_:
            if not (element in all_iv):
                return False

        return True

    # The following method is for experimental use
    def estimate_hidden_cofounder(self, method='lr'):
        full_dag = self.causal_graph.explicit_unob_var_dag
        cg = deepcopy(self.causal_graph)
        all_nodes = full_dag.nodes
        hidden_cofounders = [node for node in all_nodes if "U" in node]
        estimated = 0

        def _compute_estimated(
            ob_ancestors, node, full_dag, hidden_ancestors,
        ):
            data_input = self.data[ob_ancestors]
            target = self.data[node]

            if method == "lr":
                reg = LR()
                reg.fit(data_input, target)
                predict = reg.predict(data_input)

            key_ = 'estimate_' + hidden_ancestors[0].lower()
            nodes_ = nx.descendants(full_dag, hidden_ancestors[0])
            self.data[key_] = target - predict
            modified_nodes = list(nodes_)

            edge_list_del = []
            for i in range(len(modified_nodes)):
                for j in range(i + 1, len(modified_nodes)):
                    edge_list_del.append(
                        (modified_nodes[i], modified_nodes[j])
                    )
                    edge_list_del.append(
                        (modified_nodes[j], modified_nodes[i])
                    )

            cg.remove_edges_from(edge_list=edge_list_del, observed=False)
            cg.add_nodes([key_])
            edge_list_add = [(key_, i) for i in modified_nodes]
            cg.add_edges_from(edge_list=edge_list_add)
            estimated = estimated + 1

        if len(hidden_cofounders) == 0:
            print('Hidden cofounder estimation done. No hidden cofounder found.')
            return

        for node in all_nodes:
            if node not in hidden_cofounders:
                ancestors = nx.ancestors(full_dag, node)
                hidden_ancestors = [
                    an for an in ancestors if an in hidden_cofounders
                ]

                if len(hidden_ancestors) == 1:
                    ob_ancestors = [
                        an for an in ancestors if an not in hidden_cofounders
                    ]
                    _check_ls = [
                        check_ancestors_chain(
                            full_dag, node, hidden_ancestors[0]
                        ) for node in ob_ancestors
                    ]
                    _true_num = sum(_check_ls)
                    if _true_num == len(ob_ancestors):
                        _compute_estimated(
                            ob_ancestors, node, full_dag, hidden_ancestors
                        )

        if estimated > 0:
            print(str(estimated) + ' hidden cofounders estimated in this turn.')
            self.estimate_hidden_cofounder(method)
        else:
            print('Hidden cofounder estimation done, and there could be ' +
                  str(len(hidden_cofounders)) + ' hidden cofounders remaining.')
            return

    def __repr__(self):
        return (f'A CausalModel for {self.causal_graph}, where the treatment is'
                f'{self.treatment} and the outcome is {self.outcome}.')



"""
See (http://proceedings.mlr.press/v70/hartford17a/hartford17a.pdf) for
reference.

To use self-defined mixture density network and outcome network, one only
needs to define new MixtureDensityNetwork and OutcomeNet and wrap them with
MDNWrapper and OutcomeNetWrapper, respectively.

"""

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim

from torch.autograd import Variable
from copy import deepcopy
from sklearn.preprocessing import OneHotEncoder
from torch.utils.data import DataLoader
from torch.distributions import Categorical, Normal, MixtureSameFamily,\
    Independent

from .utils import BatchData, convert2array, convert2tensor, \
    shapes, DiscreteOBatchData, DiscreteIOBatchData, DiscreteIBatchData
from .base_models import BaseEstModel, MLModel

# We first build the mixture density network.


class MixtureDensityNetwork(nn.Module):
    """
    See (https://publications.aston.ac.uk/id/eprint/373/1/NCRG_94_004.pdf) for
    reference.
    """

    def __init__(
        self, z_d, w_d, out_d,
        hidden_d1=128,
        hidden_d2=64,
        hidden_d3=32,
        num_gaussian=5,
        is_discrete_input=False,
        is_discrete_output=False,
        embedding_dim=None,
        stab_bound=200
    ):
        """
        Parameters
        ----------
        in_d : int
            Dimension of a single data point.

        out_d : int
            Dimension of the gaussian distribution.

        hidden_d : int
            Number of neurons in the hidden layer.

        num_gaussian : int
            Number of gaussian distributions to be mixed.
        """
        super().__init__()

        self.stab_bound = stab_bound
        self._z_d = z_d
        self._w_d = w_d
        self._out_d = out_d
        self.num_gaussian = num_gaussian
        self.is_discrete_input = is_discrete_input

        if is_discrete_input:
            assert z_d is not None, 'Please specify the dimension of the'
            'treatment vector.'

            embedding_dim = z_d if embedding_dim is None else embedding_dim
            self.embed = nn.Embedding(z_d, embedding_dim)
            in_d = int(embedding_dim + w_d)
        else:
            self.embed = nn.Identity()
            in_d = int(z_d + w_d)

        self.hidden_layer = nn.Sequential(
            nn.Linear(in_d, hidden_d1),
            nn.ELU(),
            nn.Linear(hidden_d1, hidden_d2),
            nn.ELU(),
        )

        self.pi = nn.Linear(hidden_d2, num_gaussian)
        self.sigma = nn.Linear(hidden_d2, num_gaussian * out_d)
        self.mu = nn.Linear(hidden_d2, num_gaussian * out_d)

    def forward(self, x, w):
        """
        Parameters
        ----------
        x : tensor
            Has shape of (b, in_d), where b is the batch size.

        Returns
        ----------
        pi : tensor
            Mixing coefficient with the shape (b, num_gaussian) and each
            component of pi is in the range [0, 1].

        mu : tensor
            Mean of each mixing component with the shape
            (b, num_gaussian, out_d).

        sigma : tensor
            Variance with the shape (b, num_gaussian, out_d) and each
            component of sigma is large than 0.
        """
        x = self.embed(x)
        x = torch.cat((x, w), dim=1).to(torch.float32)
        h = self.hidden_layer(x)
        # h += 1 + 1e-15
        pi = self.pi(h)
        pi = pi - pi.max(dim=1).values.reshape(-1, 1)
        pi = nn.Softmax(dim=1)(pi)
        # pi = F.gumbel_softmax(pi, hard=True, dim=1)
        mu = self.mu(h).reshape(
            -1, self.num_gaussian, self._out_d
        )
        sigma = torch.clamp(self.sigma(h), -self.stab_bound, self.stab_bound)
        sigma = torch.exp(sigma).reshape(-1, self.num_gaussian, self._out_d)
        return pi, mu, sigma

# To make the above MDN consistant to the standard machine learning models, we
# use the following wrapper to wrap it so that the methods such as fit() can
# be applied.


class MDNWrapper:
    """
    Wrapped class for MixtureDensityNetwork.

    Attributes
    ----------
    mdn : MixtureDensityNetwork

    Methods
    ----------
    loss_fn(pi, mu, sigma, y)
        Calculate the loss used for training the mdn.

    fit(X, y, device='cuda', lr=0.01, epoch=1000,
        optimizer='SGD', batch_size=128)
        Train the mdn model with data (X, y).

    predict(X, y)
        Calculate the probability P(y|X) with the trained mixture density
        network.

    sample()
        Generate samples with the mixture density network.
    """

    def __init__(self, mdn):
        """
        Parameters
        ----------
        model : MixtureDensityNetwork
        """
        super().__init__()

        self.model = deepcopy(mdn)
        self._z_d = mdn._z_d
        self._w_d = mdn._w_d
        self._out_d = mdn._out_d
        self.num_gaussian = mdn.num_gaussian
        self.is_discrete_input = mdn.is_discrete_input

    def mg(self, pi, mu, sigma):
        mix = Categorical(pi)
        comp = Independent(Normal(mu, sigma), 1)
        mg = MixtureSameFamily(mix, comp)
        return mg

    def loss_fn(self, pi, mu, sigma, y):
        """Calculate the loss used for training the mdn.

        Parameters
        ----------
        pi : tensor
            Has shape (b, num_gaussian) where b is the batch size and
            num_gaussian is the number of mixiing gaussian distributions. The
            mixing coefficient of gaussian distributions.

        mu : tensor
            Shape (b, num_gaussian, out_d) where out_d is the dimension of
            each mixed gaussian distribution.

        sigma : tensor
            Has shape (b, num_gaussian, out_d). The variance of the gaussian
            distributions.

        y : tensor
            Has shape (b, out_d).

        Returns
        ----------
        tensor
            The probability of taking value y in the probability distribution
            modeled by the mdn. Has the same shape as y.
        """
        mg = self.mg(pi, mu, sigma)
        loss = mg.log_prob(y)
        return -torch.logsumexp(loss, dim=0)

    def fit(
        self, z, w,
        target,
        device='cuda',
        lr=0.01,
        epoch=100,
        optimizer='SGD',
        batch_size=64,
        **optim_config
    ):
        """Train the mdn model with data (X, y).

        Parameters
        ----------
        X : tensor
            Has shape (b, in_d) where b is the batch size or the number of data
            points and in_d is the dimension of each data point.

        y : tensor
            Has shape (b, out_d) where out_d is the dimension of each y.

        device : str, optional. Defaults to 'cuda'.

        lr : float, optional. Defaults to 0.01.
            Learning rate.

        epoch : int, optional. Defaults to 1000.
            The number of epochs used for training.

        optimizer : str, optional. Defaults to 'SGD'.
            Currently including SGD and Adam The type of optimizer used for
            training.

        batch_size: int, optional. Defaults to 128.

        optim_config : other parameters for various optimizers.
        """
        self.model = self.model.to(device)

        if w is None:
            w = z[:, -1:-1]

        if self.is_discrete_input:
            data = DiscreteIBatchData(X=z, W=w, y=target)
        else:
            data = BatchData(X=z, W=w, y=target)
        train_loader = DataLoader(data, batch_size=batch_size)

        op_dict = {
            'SGD': optim.SGD(self.model.parameters(), lr=lr),
            'Adam': optim.Adam(self.model.parameters(), lr=lr, **optim_config)
        }
        optimizer = op_dict[optimizer]

        for e in range(epoch):
            for i, (z, w, y) in enumerate(train_loader):
                self.model.train()
                z, w, y = z.to(device), w.to(device), y.to(device)
                pi, mu, sigma = self.model(z, w)
                loss = self.loss_fn(pi, mu, sigma, y)
                optimizer.zero_grad()
                loss.backward()
                optimizer.step()
            print(f'Finished {e+1}/{epoch} epochs | current loss {loss.data}')

    def predict_log_prob(self, z, w, target):
        """Calculate the probability density P(y|X) with the trained mixture density
            network.

        Parameters
        ----------
        X : tensor
            Has shape (b, in_d) where b is the batch size and in_d is the
            dimension of each data point

        y : tensor
            Has shape (b, out_d).

        Returns
        ----------
        tensor
            The probability density p(y|X) evaluated with the trained mdn.
        """
        pi, mu, sigma = self.model(z, w)
        log_prob = self.mg(pi, mu, sigma).log_prob(target)
        return log_prob

    def predict_prob(self, z, w, target):
        return torch.exp(self.predict_log_prob(z, w, target))

    def predict_cdf(self, z, w, target):
        pi, mu, sigma = self.model(z, w)
        return self.mg(pi, mu, sigma).cdf(target)

    def _sample(self, z, w, sample_n):
        # TODO: remeber to call detach to depart from the calculatin graph
        """Generate a batch of sample according to the probability density returned by
            the MDN model.

        Parameters
        ----------
        X : tensor
            Shape (b, in_d) where b is the batch size.

        sample_num : tuple of int
            Eg., (5, ) means generating (5*b) samples.

        Returns
        ----------
        tensor
            Shape (b*sample_num, out_d).
        """
        pi, mu, sigma = self.model(z, w)
        mg = self.mg(pi, mu, sigma)
        sampled = mg.sample((sample_n, ))
        return sampled.reshape(-1, self._out_d)


class Net(nn.Module):
    # This neural network is for x_net with discrete x (treatment) and all
    # y_net. If x is continuous, use MixtureDensityNetwork.
    def __init__(self, x_d, w_d, out_d,
                 hidden_d1=256,
                 hidden_d2=512,
                 hidden_d3=256,
                 is_discrete_input=False,
                 is_discrete_output=False,
                 embedding_dim=None):
        super().__init__()

        self._x_d = x_d
        self._out_d = out_d
        self.is_discrete_input = is_discrete_input
        self.is_discrete_output = is_discrete_output

        if is_discrete_input:
            assert x_d is not None, 'Please specify the dimension of the'
            'treatment vector.'

            embedding_dim = x_d if embedding_dim is None else embedding_dim
            self.embed = nn.Embedding(x_d, embedding_dim)
            in_d = int(embedding_dim + w_d)
        else:
            self.embed = nn.Identity()
            in_d = int(x_d + w_d)

        self.fc1 = nn.Linear(in_d, hidden_d1)
        self.fc2 = nn.Linear(hidden_d1, hidden_d2)
        self.fc3 = nn.Linear(hidden_d2, hidden_d3)
        self.fc4 = nn.Linear(hidden_d3, out_d)

    def forward(self, x, w):
        # this definition should be simplified
        # For discrete treatment, x should be a vector of labels not one-hot
        # tensors
        x = self.embed(x)
        x = torch.cat((x, w), dim=1)
        x = self.fc1(x)
        x = F.relu(x)
        x = self.fc2(x)
        x = F.relu(x)
        x = self.fc3(x)
        x = F.relu(x)
        output = self.fc4(x)
        return output


class NetWrapper:
    # This wrapper is for x_net with discrete x (treatment) and all y_net
    # If x is continuous, use MDNWrapper.
    def __init__(self, net, is_y_net=False):
        self.model = deepcopy(net)
        self.is_y_net = is_y_net
        self.is_discrete_input = net.is_discrete_input
        self.is_discrete_output = net.is_discrete_output
        self._x_d = net._x_d
        self._out_d = net._out_d

    def fit(
        self, x, w,
        target,
        device='cuda',
        lr=0.01,
        epoch=500,
        optimizer='SGD',
        batch_size=64,
        **optim_config
    ):
        # TODO:tqdm
        """if is_discrete_input: transform x
        if is_discrete_output: default loss should be nn.NLLLoss


        Args:
            x (_type_): _description_
            w (_type_): _description_
            target (_type_): _description_
            device (str, optional): _description_. Defaults to 'cuda'.
            lr (float, optional): _description_. Defaults to 0.01.
            epoch (int, optional): _description_. Defaults to 500.
            optimizer (str, optional): _description_. Defaults to 'SGD'.
            batch_size (int, optional): _description_. Defaults to 64.
        """
        # move model to device
        self.model = self.model.to(device)

        # convert w to tensor([]) with shape (x.shape[0], 0) if w is None
        if w is None:
            w = x[:, -1:-1]

        # use different data set for efficiency
        if self.is_discrete_output:
            if self.is_discrete_input and not self.is_y_net:
                data = DiscreteIOBatchData(X=x, W=w, y=target)
            else:
                data = DiscreteOBatchData(X=x, W=w, y=target)

            loss_fn = optim_config.pop('loss', nn.CrossEntropyLoss())
        else:
            if self.is_discrete_input and not self.is_y_net:
                data = DiscreteIBatchData(X=x, W=w, y=target)
            else:
                data = BatchData(X=x, W=w, y=target)

            loss_fn = optim_config.pop('loss', nn.MSELoss())

        # make batched data
        train_loader = DataLoader(data, batch_size=batch_size)

        op_dict = {
            'SGD': optim.SGD(self.model.parameters(), lr=lr),
            'Adam': optim.Adam(self.model.parameters(), lr=lr, **optim_config)
        }
        # prepare for training
        optimizer = op_dict[optimizer]

        if self.is_discrete_input and self.is_y_net:
            x_d = self._x_d
            out_d = self._out_d
            x_label = torch.arange(x_d)
            for e in range(epoch):
                for i, (X, W, y) in enumerate(train_loader):
                    self.model.train()
                    X, W, y = X.to(device), W.to(device), y.to(device)

                    x = Variable(X)  # TODO: I dont konw why this works, why???
                    batch_num = y.shape[0]
                    # after sampled, both batches of w_sample and x_sample
                    # will have (batch_num*x_d) samples
                    w_sample = W.repeat_interleave(x_d, dim=0)
                    x_sample = x_label.repeat(batch_num, )

                    # y_pred_vec has shape (batch_num*x_d, out_d)
                    y_pred = self.model(x_sample, w_sample)
                    y_pred = y_pred.reshape(batch_num, x_d, out_d)
                    y_pred = torch.einsum('bkj,bk->bj', [y_pred, x])
                    loss = loss_fn(y_pred, y)

                    optimizer.zero_grad()
                    # loss.backward(retain_graph=True)
                    loss.backward()
                    optimizer.step()
                print(
                    f'Finished {e+1}/{epoch} epochs | current loss {loss.data}'
                )
        else:
            for e in range(epoch):
                for i, (X, W, y) in enumerate(train_loader):
                    self.model.train()
                    X, W, y = X.to(device), W.to(device), y.to(device)
                    y_pred = self.model(X, W)
                    loss = loss_fn(y_pred, y)

                    optimizer.zero_grad()
                    loss.backward()
                    optimizer.step()
                print(
                    f'Finished {e+1}/{epoch} epochs | current loss {loss.data}'
                )

    def predict_proba(self, x, w):
        assert self.is_discrete_output, 'Please use predict(x, w) for'
        'continous output.'
        if self.is_discrete_input:
            x = torch.argmax(x, dim=1)

        pred = self.model(x, w)
        return nn.Softmax(dim=1)(pred)

    def predict(self, x, w):
        if self.is_discrete_input:
            x = torch.argmax(x, dim=1)

        if self.is_discrete_output:
            pred = nn.Softmax(dim=1)(self.model(x, w))
            return torch.argmax(pred, dim=1)
        else:
            return self.model(x, w)

    def _sample(self, x, w, sample_n):
        return self.predict_proba(x, w)


# We are now ready to build the complete model with above wrapped outcome and
# treatment nets.


class DeepIV(BaseEstModel):
    r"""Training of a DeepIV model g(x, w) is composed of 2 stages:
            1. In the first stage, we train a neural network to estimate the
            distribution of the treatment x given the instrument z and
            adjustment (also covariate) w;
            2. In the second stage, we train another neural network to estiamte
            the outcome y givn treatment x and adjustment (also
            covariate) w.
        The trained model is used to estimate the causal effect
            g(x_1, w) - g(x_0, w)
        or
            \partial_x g(x, w).

    Attributes
    ----------
    x_net_kwargs : dict

    y_net_kwargs : dict

    _x_net_init : Net
        The model for the treatment.

    _y_net_init : Net
        The model for the outcome.

    x_net : NetWrapper, optional
        Wrapped treatment model.

    y_net : NetWrapper, optional
        Wrapped outcome model.

    is_discrete_instrument : bool
        True if the instrument is discrete.

    randome_state : int

    is_discrete_treatment : bool

    is_discrete_outcome : bool

    categories : str, optional. Defaults to 'auto'

    x_transformer : OneHotEncoder, optional
        Transformer of the treatment.

    _z_transformer : OneHotEncoder, optional
        Transformer of the instrument

    y_transformer : OneHotEncoder, optional
        Transformer of the outcome

    w : ndarray with shape (n, w_d)
        Adjustment variables in the training data where n is the number of
        examples and w_d is the number of adjustment and covaraite.

    _y_d : ndarray
        Shapes of the outcome variables in the training data.

    _x_d : ndarray
        Shapes of the treatment variables in the training data.

    _w_d : ndarray
        Shapes of the adjustment variables in the training data.

    _z_d : ndarray
        Shapes of the instrumental variables in the training data.

    Methods
    ----------
    fit(data, outcome, treatment, instrument=None, adjustment=None,
        approx_grad=True, sample_n=None, x_net_config=None, 
        y_net_config=None, **kwargs)
        Fit the instance of DeepIV.

    estimate(data=None, treat=None, control=None, quantity='CATE', marginal_effect=False,
             *args, **kwargs,)
        Estimate the causal effect.

    _gen_x_model(x_model, *args, **kwargs)

    _gen_y_model(y_model, *args, **kwargs)

    _prepare4est(data, treat=None, control=None, marginal_effect=False)

    """

    def __init__(
        self,
        x_net=None,
        y_net=None,
        x_hidden_d=None,
        y_hidden_d=None,
        num_gaussian=5,
        is_discrete_treatment=False,
        is_discrete_outcome=False, 
        is_discrete_instrument=False,
        categories='auto',
        random_state=2022,
    ):
        """
        Parameters
        ----------
        x_net : ylearn.estimator_model.deepiv.Net
            Representation of the mixture density network for continuous
            treatment or an usual classification net for discrete treatment. If None, the default neural network will be used. See :py:class:`ylearn.estimator_model.deepiv.Net` for reference.

        y_net :  ylearn.estimator_model.deepiv.Net
            Representation of the outcome network. If None, the default neural network will be used. See :py:class:`ylearn.estimator_model.deepiv.Net` for reference.

        x_hidden_d : int, optional. Defaults to None
            Dimension of the hidden layer of the default x_net of DeepIV.

        y_hidden_d : int, optional. Defaults to None
            Dimension of the hidden layer of the default y_net of DeepIV.

        num_gaussian : int, optional. Defaults to 5.
            Number of gaussians when using the mixture density network which will be directly ignored when the treatment is discrete.

        is_discrete_treatment : bool, optional. Defaults to False.

        is_discrete_outcome : bool, optional. Defaults to False.

        is_discrete_instrument : bool, optional. Defaults to False.

        categories : str, optional. Defaults to 'auto'

        random_state : int, optional. Defaults to 2022.
        """
        self.x_net_kwargs = {}
        if x_hidden_d is not None:
            self.x_net_kwargs['hidden_d'] = x_hidden_d
        if num_gaussian is not None and not is_discrete_treatment:
            self.x_net_kwargs['num_gaussian'] = num_gaussian

        self.y_net_kwargs = {}
        if y_hidden_d is not None:
            self.y_net_kwargs['hidden_d'] = y_hidden_d

        self._x_net_init = x_net
        self._y_net_init = y_net
        self.is_discrete_instrument = is_discrete_instrument

        super().__init__(
            random_state=random_state,
            is_discrete_treatment=is_discrete_treatment,
            is_discrete_outcome=is_discrete_outcome,
            # is_discrete_instrument=is_discrete_instrument,
            categories=categories,
        )

    def fit(
        self,
        data,
        outcome,
        treatment,
        instrument=None,
        adjustment=None,
        approx_grad=True,
        sample_n=None,
        x_net_config=None,
        y_net_config=None,
        **kwargs
    ):
        """Train the DeepIV model. #TODO: consider implementing the comp_transformer for multiple treatment

        Parameters
        ----------
        data : DataFrame
            Training dataset for training the estimator.
        outcome : list of str, optional
            Names of the outcome
        treatment : list of str, optional
            Names of the treatment
        instrument : list of str, optional
            Names of the instrument, by default None
        adjustment : list of str, optional
            Names of the adjustment set. #TODO: Note that in the current version we aslo view all adjustment variables as the covariates., by default None
        approx_grad : bool, optional
            Whether use the approximated gradient as in the reference, by default True
        sample_n : int, optional
            Times of new samples when using the approx_grad technique, by default None
        x_net_config : dict, optional
            Configuration of the x_net, by default None
        y_net_config : dict, optional
            Configuration of the y_net, by default None

        Returns
        -------
        instance of DeepIV
            The trained DeepIV model
        """
        assert instrument is not None, 'instrument is required.'

        super().fit(data, outcome, treatment,
                    adjustment=adjustment,
                    instrument=instrument,
                    )

        if x_net_config is None and y_net_config is None:
            x_net_config = kwargs
            y_net_config = kwargs

        y, x, z, w = convert2array(
            data, outcome, treatment, instrument, adjustment
        )

        # transformers for building one hot vectors
        if self.is_discrete_treatment:
            if self.categories == 'auto' or self.categories is None:
                categories = 'auto'
            else:
                categories = list(self.categories)

            self.x_transformer = OneHotEncoder(categories=categories)
            self.x_transformer.fit(x)
            x = self.x_transformer.transform(x).toarray()

        if self.is_discrete_instrument:
            self._z_transformer = OneHotEncoder()
            self._z_transformer.fit(z)
            z = self._z_transformer.transform(z).toarray()

        if self.is_discrete_outcome:
            self.y_transformer = OneHotEncoder()
            self.y_transformer.fit(y)
            y = self.y_transformer.transform(y).toarray()

        y, x, w, z = convert2tensor(y, x, w, z)
        self.w = w
        self._y_d, self._x_d, self._w_d, self._z_d = shapes(
            y, x, w, z, all_dim=False
        )

        # build networks
        self.x_net = self._gen_x_model(
            self._x_net_init,
            self._z_d,
            self._w_d,
            self._x_d,
            is_discrete_input=self.is_discrete_instrument,
            is_discrete_output=self.is_discrete_treatment,
            **self.x_net_kwargs
        )
        self.y_net = self._gen_y_model(
            self._y_net_init,
            self._x_d,
            self._w_d,
            self._y_d,
            is_discrete_input=self.is_discrete_treatment,
            is_discrete_output=self.is_discrete_outcome,
            **self.y_net_kwargs
        )

        # Step 1: train the model for estimating the treatment given the
        # instrument and adjustment
        self.x_net.fit(z, w, target=x, **x_net_config)

        # Step 2: generate new samples if calculating grad approximately
        if approx_grad:
            x_sampled = self.x_net._sample(z, w, sample_n)
        else:
            # TODO: the loss funcn should be modified if not approx_grad
            x_sampled = x

        # Step 3: fit the final counterfactual prediction model
        if self.is_discrete_treatment:
            w_sampled = w
        else:
            w_sampled = w.repeat(sample_n, 1) if w is not None else None

        # TODO: be careful here
        self.y_net.fit(x_sampled, w_sampled, target=y, **y_net_config)

        self._is_fitted = True
        return self

    def effect_nji(self, data=None):
        if not hasattr(self, 'x_net') or not hasattr(self, 'y_net'):
            raise Exception('The estimator is not fitted yet.')
        
        if data is None:
            w = self.w
        else:
            w = convert2tensor(
                convert2array(data, self.adjustment)[0]
            )[0]

        n = w.shape[0]
        ones = torch.eye(n, self._x_d)

        y_nji = torch.full((n, self._y_d, self._x_d))

        if self.is_discrete_treatment:
            for treat in range(self._x_d):
                treat_id = (torch.ones(n, ) * treat).int()

                xt = ones.index_select(dim=0, index=treat_id)
                y_pred = self.y_net.predict(xt, w)
                y_nji[:, :, treat] = y_pred.reshape(n, self._y_d)

            y_ctrl = y_nji[:, :, 0].reshape(n, -1, 1).repeat((1, 1, self._x_d))
        else:
            xt = ones * 1
            x0 = ones * 0
            y_nji[:, :, 0] = self.y_net.predict(x0, w)
            y_nji[:, :, 1] = self.y_net.predict(xt, w)
            y_ctrl = y_nji[:, :, 0].reshape(n, -1, 1).repeat((1, 1, 2))
        
        y_nji = y_nji - y_ctrl
        
        return y_nji

    def _prepare4est(
        self,
        data=None,
        treat=None,
        control=None,
        marginal_effect=False,
        *args,
        **kwargs
    ):
        if not hasattr(self, 'x_net') or not hasattr(self, 'y_net'):
            raise Exception('The estimator is not fitted yet.')

        treat = 1 if treat is None else treat
        control = 0 if control is None else control

        if data is None:
            w = self.w
        else:
            w = convert2tensor(
                convert2array(data, self.adjustment)[0]
            )[0]
        n = w.shape[0]
        ones = torch.eye(n, self._x_d)

        if self.is_discrete_treatment:
            # build the one_hot vector for treatment vector
            treat_id = (torch.ones(n, ) * treat).int()

            # build treatment vector xt and control vector x0
            xt = ones.index_select(dim=0, index=treat_id)
            x0 = ones.index_select(dim=0, index=torch.zeros(n, ).int())

            return (self.y_net.predict(xt, w), self.y_net.predict(x0, w))
        else:
            xt = ones * treat
            x0 = ones * control
            xt.requires_grad = True

            yt = self.y_net.predict(xt, w)

            if marginal_effect:
                return (xt.grad.detach(), )
            else:
                y0 = self.y_net.predict(x0, w)
                return (yt, y0)

    def estimate(
        self,
        data=None,
        treat=None,
        control=None,
        quantity='CATE',
        marginal_effect=False,
        *args,
        **kwargs,
    ):
        #TODO: update the definition of treat and control.
        """Estimate the causal effect with the type of the quantity.

        Parameters
        ----------
        data : DataFrame, optional
            Test data. The model will use the training data if None, by default None
        treat : int, optional
            Value of the treatment, by default None. If None, then the model will set treat=1.
        control : int, optional
            Value of the control, by default None. If None, then the model will set control=0.
        quantity : str, optional
            Option for returned estimation result. The possible values of quantity include:
                
                1. *'CATE'* : the estimator will evaluate the CATE;
                
                2. *'ATE'* : the estimator will evaluate the ATE;
                
                3. *None* : the estimator will evaluate the ITE or CITE.
        marginal_effect : bool, optional
            _description_, by default False

        Returns
        -------
        torch.tensor
            Estimated causal effects
        """
        y_preds = self._prepare4est(
            data=data,
            treat=treat,
            control=control,
            marginal_effect=marginal_effect,
            *args,
            **kwargs
        )

        if not marginal_effect:
            yt, y0 = y_preds
        else:
            yt, y0 = y_preds[0], None

        if quantity == 'CATE' or quantity == 'ATE':
            return (yt - y0).mean(dim=0) if y0 is not None else yt.mean(dim=0)
        # elif quantity == 'Counterfactual prediction':
        else:
            return yt

    def _gen_x_model(self, x_model, *args, **kwargs):
        hidden_d_list = kwargs.pop('hidden_d', [64, 128, 64])
        for i, hidden_d in enumerate(hidden_d_list, 1):
            kwargs[f'hidden_d{i}'] = hidden_d

        if self.is_discrete_treatment:
            if x_model is None:
                assert any(args), 'Need parameters to define a treatment net.'
                x_model = Net(*args, **kwargs)

            x_net = NetWrapper(x_model, is_y_net=False)
        else:
            if x_model is None:
                assert any(args), 'Need parameters to define a treatment net.'
                x_model = MixtureDensityNetwork(*args, **kwargs)

            x_net = MDNWrapper(x_model)

        return x_net

    def _gen_y_model(self, y_model, *args, **kwargs):
        hidden_d_list = kwargs.pop('hidden_d', [64, 128, 64])
        for i, hidden_d in enumerate(hidden_d_list, 1):
            kwargs[f'hidden_d{i}'] = hidden_d

        if y_model is None:
            assert any(args), 'Need parameters to define an outcome net.'
            y_net = Net(*args, **kwargs)

        return NetWrapper(y_net, is_y_net=True)



# -*- coding:utf-8 -*-
from __future__ import absolute_import

import os
import sys
from glob import glob
from os import path as P

import numpy
from setuptools import setup, Extension, find_packages

my_name = 'ylearn'


def read_requirements(file_path='requirements.txt'):
    if not os.path.exists(file_path):
        return []

    with open(file_path, 'r') as f:
        lines = f.readlines()

    lines = [x.strip('\n').strip(' ') for x in lines]
    lines = list(filter(lambda x: len(x) > 0 and not x.startswith('#'), lines))

    return lines


def read_extra_requirements():
    import glob
    import re

    extra = {}

    for file_name in glob.glob('requirements-*.txt'):
        key = re.search('requirements-(.+).txt', file_name).group(1)
        req = read_requirements(file_name)
        if req:
            extra[key] = req

    if extra and 'all' not in extra.keys():
        extra['all'] = sorted({v for req in extra.values() for v in req})

    return extra


try:
    execfile
except NameError:
    def execfile(fname, globs, locs=None):
        locs = locs or globs
        exec(compile(open(fname).read(), fname, "exec"), globs, locs)

HERE = P.dirname((P.abspath(__file__)))
version_ns = {}
execfile(P.join(HERE, my_name, '_version.py'), version_ns)
version = version_ns['__version__']
print("__version__=" + version)

np_include = numpy.get_include()
pyx_files = glob(f"{my_name}/**/*.pyx", recursive=True)
c_files = glob(f"{my_name}/**/*.cpp", recursive=True)
build_ext = any(map(lambda s: s == 'build_ext', sys.argv[1:]))

if build_ext:
    pyx_modules = [P.splitext(f)[0] for f in pyx_files]
    c_modules = [P.splitext(f)[0] for f in c_files]
    c_modules = [f for f in c_modules if f not in pyx_modules]
else:
    c_modules = [P.splitext(f)[0] for f in c_files]
    for pf in pyx_files:
        if P.splitext(pf)[0] not in c_modules:
            raise FileNotFoundError(f'Not found c file for {pf}, '
                                    f'run "python setup.py build_ext --inplace" to generate c files.')
    pyx_modules = []
print('pyx extensions:', pyx_modules)
print('cpp extensions:', c_modules)
print('np_include', np_include)

c_modules = list(map(lambda f: Extension(f.replace(os.sep, '.'), [f'{f}.cpp'], include_dirs=[np_include]), c_modules))
if pyx_modules:
    from Cython.Build import cythonize

    pyx_modules = list(map(lambda f: Extension(f.replace(os.sep, '.'), [f'{f}.pyx'],
                                               include_dirs=[np_include],
                                               language="c++"),
                           pyx_modules))
    pyx_modules = cythonize(pyx_modules, compiler_directives={'language_level': "3"})

MIN_PYTHON_VERSION = '>=3.6.*'

long_description = open('README.md', encoding='utf-8').read()

requires = read_requirements()
extras_require = read_extra_requirements()
print('requirements:', requires)

setup(
    name='myylearn',
    version=version,
    description='A python package for causal inference',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/DataCanvasIO/YLearn',
    author='DataCanvas Community',
    author_email='yangjian@zetyun.com',
    license='Apache License 2.0',
    install_requires=requires,
    python_requires=MIN_PYTHON_VERSION,
    extras_require=extras_require,
    classifiers=[
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    packages=find_packages(exclude=('docs', 'tests', 'example_usages')),
    package_data={
        'ylearn': ['*.txt', '**/**.txt', ],
    },
    ext_modules=c_modules + pyx_modules,
    include_dirs=[np_include],
    zip_safe=False,
    # entry_points={
    #     # 'console_scripts': [
    #     #     'why = ylearn.foo:main',
    #     # ],
    # },
    include_package_data=True,
)

# setup extensions:
#     python setup.py build_ext --inplace



import copy
import math

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator
from sklearn.compose import make_column_selector
from sklearn.impute import SimpleImputer
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import OrdinalEncoder, LabelEncoder, StandardScaler

from ylearn.utils import logging, const, infer_task_type
from ._data_cleaner import DataCleaner
from ._dataframe_mapper import DataFrameMapper

logger = logging.get_logger(__name__)


class SafeOrdinalEncoder(OrdinalEncoder):
    __doc__ = r'Adapted from sklearn OrdinalEncoder\n' + OrdinalEncoder.__doc__

    def transform(self, X, y=None):
        if not isinstance(X, (pd.DataFrame, np.ndarray)):
            raise TypeError("Unexpected type {}".format(type(X)))

        def make_encoder(categories):
            unseen = len(categories)
            m = dict(zip(categories, range(unseen)))
            vf = np.vectorize(lambda x: m[x] if x in m.keys() else unseen)
            return vf

        values = X if isinstance(X, np.ndarray) else X.values
        encoders_ = [make_encoder(cat) for cat in self.categories_]
        result = [encoders_[i](values[:, i]) for i in range(values.shape[1])]

        if isinstance(X, pd.DataFrame):
            assert len(result) == len(X.columns)
            data = {c: result[i] for i, c in enumerate(X.columns)}
            result = pd.DataFrame(data, dtype=self.dtype)
        else:
            result = np.stack(result, axis=1)
            if self.dtype != result.dtype:
                result = result.astype(self.dtype)

        return result

    def inverse_transform(self, X):
        if not isinstance(X, (pd.DataFrame, np.ndarray)):
            raise TypeError("Unexpected type {}".format(type(X)))

        def make_decoder(categories, dtype):
            if dtype in (np.float32, np.float64, float):
                default_value = np.nan
            elif dtype in (np.int32, np.int64, np.uint32, np.uint64, np.uint, int):
                default_value = -1
            else:
                default_value = None
                dtype = object
            unseen = len(categories)
            vf = np.vectorize(lambda x: categories[x] if unseen > x >= 0 else default_value,
                              otypes=[dtype])
            return vf

        values = X if isinstance(X, np.ndarray) else X.values
        decoders_ = [make_decoder(cat, cat.dtype) for i, cat in enumerate(self.categories_)]
        result = [decoders_[i](values[:, i]) for i in range(values.shape[1])]

        if isinstance(X, pd.DataFrame):
            assert len(result) == len(X.columns)
            data = {c: result[i] for i, c in enumerate(X.columns)}
            result = pd.DataFrame(data)
        else:
            result = np.stack(result, axis=1)

        return result


class FeatureImportancesSelectionTransformer(BaseEstimator):
    STRATEGY_THRESHOLD = 'threshold'
    STRATEGY_QUANTILE = 'quantile'
    STRATEGY_NUMBER = 'number'

    _default_strategy = dict(
        default_strategy=STRATEGY_THRESHOLD,
        default_threshold=0.1,
        default_quantile=0.2,
        default_number=0.8,
    )

    def __init__(self, task=None, strategy=None, threshold=None, quantile=None, number=None, data_clean=True):
        super().__init__()

        self.task = task
        self.strategy = strategy
        self.threshold = threshold
        self.quantile = quantile
        self.number = number
        self.data_clean = data_clean

        # fitted
        self.feature_names_in_ = None
        self.n_features_in_ = None
        self.feature_importances_ = None
        self.selected_features_ = None

    def fit(self, X, y):
        if self.task is None:
            self.task, _ = infer_task_type(y)

        columns_in = X.columns.to_list()
        # logger.info(f'all columns: {columns}')

        if self.data_clean:
            logger.info('data cleaning')
            kwargs = dict(replace_inf_values=np.nan, drop_label_nan_rows=True,
                          drop_constant_columns=True, drop_duplicated_columns=False,
                          drop_idness_columns=True, reduce_mem_usage=False,
                          correct_object_dtype=False, int_convert_to=None,
                          )
            dc = DataCleaner(**kwargs)
            X, y = dc.fit_transform(X, y)
            assert set(X.columns.tolist()).issubset(set(columns_in))

        preprocessor = general_preprocessor()
        estimator = general_estimator(X, y, task=self.task)

        if self.task != 'regression' and y.dtype != 'int':
            logger.info('label encoding')
            le = LabelEncoder()
            y = le.fit_transform(y)

        logger.info('preprocessing')
        X = preprocessor.fit_transform(X, y)
        logger.info('scoring')
        estimator.fit(X, y)
        importances = estimator.feature_importances_

        selected, unselected = \
            self._select_feature_by_importance(importances, strategy=self.strategy,
                                               threshold=self.threshold,
                                               quantile=self.quantile,
                                               number=self.number)
        columns = X.columns.to_list()
        selected = [columns[i] for i in selected]

        if len(columns) != len(columns_in):
            importances = [0.0 if c not in columns else importances[columns.index(c)] for c in columns_in]
            importances = np.array(importances)

        self.n_features_in_ = len(columns_in)
        self.feature_names_in_ = columns_in
        self.feature_importances_ = importances
        self.selected_features_ = selected

        # logger.info(f'selected columns:{self.selected_features_}')

        return self

    def transform(self, X):
        return X[self.selected_features_]

    @classmethod
    def _detect_strategy(cls, strategy, *, threshold=None, quantile=None, number=None,
                         default_strategy, default_threshold, default_quantile, default_number):
        if strategy is None:
            if threshold is not None:
                strategy = cls.STRATEGY_THRESHOLD
            elif number is not None:
                strategy = cls.STRATEGY_NUMBER
            elif quantile is not None:
                strategy = cls.STRATEGY_QUANTILE
            else:
                strategy = default_strategy

        if strategy == cls.STRATEGY_THRESHOLD:
            if threshold is None:
                threshold = default_threshold
        elif strategy == cls.STRATEGY_NUMBER:
            if number is None:
                number = default_number
        elif strategy == cls.STRATEGY_QUANTILE:
            if quantile is None:
                quantile = default_quantile
            assert 0 < quantile < 1.0
        else:
            raise ValueError(f'Unsupported strategy: {strategy}')

        return strategy, threshold, quantile, number

    @classmethod
    def _select_feature_by_importance(cls, feature_importance,
                                      strategy=None, threshold=None, quantile=None, number=None):
        assert isinstance(feature_importance, (list, tuple, np.ndarray)) and len(feature_importance) > 0

        strategy, threshold, quantile, number = cls._detect_strategy(
            strategy, threshold=threshold, quantile=quantile, number=number,
            **cls._default_strategy)

        feature_importance = np.array(feature_importance)
        idx = np.arange(len(feature_importance))

        if strategy == cls.STRATEGY_THRESHOLD:
            selected = np.where(np.where(feature_importance >= threshold, idx, -1) >= 0)[0]
        elif strategy == cls.STRATEGY_QUANTILE:
            q = np.quantile(feature_importance, quantile)
            selected = np.where(np.where(feature_importance >= q, idx, -1) >= 0)[0]
        elif strategy == cls.STRATEGY_NUMBER:
            if isinstance(number, float) and 0 < number < 1.0:
                number = math.ceil(len(feature_importance) * number)
            pos = len(feature_importance) - number
            sorted_ = np.argsort(np.argsort(feature_importance))
            selected = np.where(sorted_ >= pos)[0]
        else:
            raise ValueError(f'Unsupported strategy: {strategy}')

        unselected = list(set(range(len(feature_importance))) - set(selected))
        unselected = np.array(unselected)

        return selected, unselected


def general_preprocessor(*, number_scaler=None):
    cat_steps = [('imputer_cat', SimpleImputer(strategy='constant', fill_value='')),
                 ('encoder', SafeOrdinalEncoder())]
    num_steps = [('imputer_num', SimpleImputer(strategy='mean')),
                 # ('scaler', StandardScaler()),
                 ]
    if number_scaler is True:
        num_steps.append(('scaler', StandardScaler()))

    cat_transformer = Pipeline(steps=cat_steps)
    num_transformer = Pipeline(steps=num_steps)
    bool_transformer = SafeOrdinalEncoder()

    cat_selector = make_column_selector(dtype_include=['object', 'category'])
    bool_selector = make_column_selector(dtype_include='bool')
    num_selector = make_column_selector(dtype_include='number', dtype_exclude='timedelta')

    preprocessor = DataFrameMapper(
        features=[(cat_selector, cat_transformer),
                  (bool_selector, bool_transformer),
                  (num_selector, num_transformer)],
        input_df=True,
        df_out=True)
    return preprocessor


def general_estimator(X, y=None, estimator=None, task=None, random_state=None, **kwargs):
    try:
        import lightgbm
        lightgbm_installed = True
    except ImportError:
        lightgbm_installed = False

    def default_gbm(task_):
        est_cls = lightgbm.LGBMRegressor if task_ == const.TASK_REGRESSION else lightgbm.LGBMClassifier
        options = dict(n_estimators=50,
                       num_leaves=15,
                       max_depth=5,
                       subsample=0.5,
                       subsample_freq=1,
                       colsample_bytree=0.8,
                       reg_alpha=1,
                       reg_lambda=1,
                       importance_type='gain',
                       random_state=random_state,
                       verbose=-1,
                       **kwargs
                       )
        return est_cls(**options)

    def default_dt(task_):
        from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor
        est_cls = DecisionTreeRegressor if task_ == const.TASK_REGRESSION else DecisionTreeClassifier
        options = dict(min_samples_leaf=20, min_impurity_decrease=0.01, random_state=random_state, **kwargs)
        return est_cls(**options)

    def default_rf(task_):
        from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
        est_cls = RandomForestRegressor if task_ == const.TASK_REGRESSION else RandomForestClassifier
        options = dict(min_samples_leaf=20, min_impurity_decrease=0.01, random_state=random_state, **kwargs)
        return est_cls(**options)

    def default_gb(task_):
        from sklearn.ensemble import GradientBoostingClassifier, GradientBoostingRegressor
        est_cls = GradientBoostingRegressor if task_ == const.TASK_REGRESSION else GradientBoostingClassifier
        options = dict(n_estimators=100, max_depth=100, random_state=random_state, **kwargs)
        return est_cls(**options)

    def default_lr(task_):
        from sklearn.linear_model import LinearRegression, LogisticRegression

        if task_ == const.TASK_REGRESSION:
            est_cls = LinearRegression
            options = kwargs
        else:
            est_cls = LogisticRegression
            options = dict(random_state=random_state, **kwargs)
        return est_cls(**options)

    def default_lasso(task_):
        assert task_ == const.TASK_REGRESSION
        from sklearn.linear_model import Lasso
        options = dict(random_state=random_state, **kwargs)
        return Lasso(**options)

    creators = dict(
        gbm=default_gbm,
        lgbm=default_gbm,
        dt=default_dt,
        rf=default_rf,
        gb=default_gb,
        lr=default_lr,
        lasso=default_lasso,
    )

    if estimator is None:
        estimator = 'gbm' if lightgbm_installed else 'rf'
    if task is None:
        assert y is not None, '"y" or "task" is required.'
        task = infer_task_type(y)
    elif isinstance(task, bool):  # discrete or not
        task = const.TASK_MULTICLASS if task else const.TASK_REGRESSION

    if estimator in creators.keys():
        estimator_ = creators[estimator](task)
    else:
        estimator_ = copy.deepcopy(estimator)

    return estimator_



import copy
from itertools import product

import numpy as np
import pandas as pd
from joblib import delayed, Parallel

from .base_models import BaseEstModel
from .doubly_robust import DoublyRobust
from .meta_learner import SLearner, TLearner, XLearner


def _copy_and_fit(learner, data, outcome, treatment, treat, control, **kwargs):
    learner = copy.deepcopy(learner)
    learner.fit(data, outcome, treatment, treat=treat, control=control, **kwargs)
    return learner


def _effect_nji(learner_and_sign, data, outcome, treatment, treat, control, ):
    y_count = 1 if isinstance(outcome, str) else len(outcome)

    if treat == control:
        effect = np.zeros((len(data), y_count, y_count))
    else:
        assert isinstance(learner_and_sign, tuple)

        learner, sign = learner_and_sign
        effect = learner.effect_nji(data)
        assert effect.shape[1] == y_count
        assert effect.shape[2] == 2

        if np.all(effect[:, :, 1:] == 0.0):
            effect = effect[:, :, :1]
        else:
            effect = effect[:, :, 1:]
        if sign < 0:
            effect = effect * sign

    return effect


class PermutedLearner(BaseEstModel):
    def __init__(self, learner):
        assert isinstance(learner, BaseEstModel) and learner.is_discrete_treatment
        assert not learner._is_fitted

        super().__init__(random_state=learner.random_state,
                         is_discrete_treatment=learner.is_discrete_treatment,
                         is_discrete_outcome=learner.is_discrete_outcome,
                         _is_fitted=False,
                         # is_discrete_instrument=False,
                         categories=learner.categories)

        self.learner = learner

        # fitted
        self.learners_ = {}
        self._is_fitted = False

        # others
        self._last_treat_control = None

    def _permute_treats(self):
        treatment = self.treatment
        if isinstance(treatment, str):
            treatment = [treatment]
        assert isinstance(treatment, (list, tuple)) and len(treatment) > 0

        single_treatment = len(treatment) == 1
        if single_treatment:
            permuted_treats = self.treats_[treatment[0]]
        else:
            permuted_treats = [self.treats_[t] for t in treatment]
            permuted_treats = tuple(product(*permuted_treats))

        n = len(permuted_treats)
        for t in range(1, n):
            for c in range(t):
                if single_treatment:
                    yield permuted_treats[t], permuted_treats[c]
                else:
                    yield tuple(permuted_treats[t]), tuple(permuted_treats[c])

    def _get_learner(self, treat, control, silent=False):
        assert self._is_fitted

        treatment = self.treatment
        if isinstance(treatment, str):
            treatment = [treatment]
        assert isinstance(treatment, (list, tuple)) and len(treatment) > 0

        single_treatment = len(treatment) == 1

        if treat is None:
            treat = [self.treats_[t][-1] for t in treatment]
            if single_treatment:
                treat = treat[0]

        if control is None:
            control = [self.treats_[t][0] for t in treatment]
            if single_treatment:
                control = control[0]

        if isinstance(treat, list):
            treat = tuple(treat)
        if isinstance(control, list):
            control = tuple(control)

        if (treat, control) in self.learners_.keys():
            return self.learners_[(treat, control)], 1
        elif (control, treat) in self.learners_.keys():
            return self.learners_[(control, treat)], -1
        else:
            if not silent:
                raise ValueError(f'Not found leaner for treat-control pair: [{treat},{control}]')
            else:
                return None

    def fit(
            self,
            data,
            outcome,
            treatment,
            # adjustment=None,
            # covariate=None,
            n_jobs=None,
            **kwargs,
    ):
        assert self.is_discrete_treatment
        assert isinstance(data, pd.DataFrame)

        super().fit(data, outcome, treatment, **kwargs)

        learners = {}
        if n_jobs in {0, 1}:
            # fit learners one by one
            for treat, control in self._permute_treats():
                learners[(treat, control)] = _copy_and_fit(
                    self.learner, data, outcome, treatment, treat, control, **kwargs
                )
        else:
            # fit learners with joblib
            job_options = self._get_job_options(n_jobs)
            tc_pairs = list(self._permute_treats())
            ls = Parallel(**job_options)(delayed(_copy_and_fit)(
                self.learner, data, outcome, treatment, t, c, **kwargs
            ) for t, c in tc_pairs)
            for (t, c), l in zip(tc_pairs, ls):
                learners[(t, c)] = l

        self.learners_ = learners
        self._is_fitted = True
        return self

    def estimate(self, data=None, treat=None, control=None, **kwargs):
        learner, sign = self._get_learner(treat, control)
        effect = learner.estimate(data, **kwargs)
        if sign < 0:
            effect = effect * sign

        self._last_treat_control = (treat, control)
        return effect

    def effect_nji(self, data=None, control=None, n_jobs=None, **kwargs):
        treatment = self.treatment
        if isinstance(treatment, str):
            treatment = [treatment]

        if control is None:
            control = [self.treats_[t][0] for t in treatment]

        if isinstance(control, list):
            control = tuple(control)

        if len(treatment) == 1:
            control = control[0]
            treats = self.treats_[treatment[0]]
        else:
            treats = product(*[self.treats_[x] for x in treatment])

        if n_jobs in {0, 1}:
            # estimate learners one by one
            effects = [_effect_nji(
                self._get_learner(treat, control, silent=True),
                data, self.outcome, self.treatment, treat, control)
                for treat in treats]
        else:
            # estimate learners with joblib
            job_options = self._get_job_options(n_jobs)
            effects = Parallel(**job_options)(delayed(_effect_nji)(
                self._get_learner(treat, control, silent=True),
                data, self.outcome, self.treatment, treat, control
            ) for treat in treats)

        effects = np.concatenate(effects, axis=2)
        return effects

    @staticmethod
    def _get_job_options(n_jobs=None):
        return dict(n_jobs=n_jobs, prefer='processes')

    @property
    def treat(self):
        tc = self._last_treat_control
        if tc is None:
            raise ValueError('Not found the last_treat_control, call estimate firstly please.')

        learner, _ = self._get_learner(tc[0], tc[1])
        return learner.treat

    def __repr__(self):
        return f'Permuted{self.learner}'


class PermutedSLearner(PermutedLearner):
    def __init__(self, model, *args, **kwargs):
        learner = SLearner(model, *args, **kwargs)
        super().__init__(learner)


class PermutedTLearner(PermutedLearner):
    def __init__(self, model, *args, **kwargs):
        learner = TLearner(model, *args, **kwargs)
        super().__init__(learner)


class PermutedXLearner(PermutedLearner):
    def __init__(self, model, *args, **kwargs):
        learner = XLearner(model, *args, **kwargs)
        super().__init__(learner)


class PermutedDoublyRobust(PermutedLearner):
    def __init__(self, x_model, y_model, yx_model, *args, **kwargs):
        learner = DoublyRobust(x_model, y_model, yx_model, *args, **kwargs)
        super().__init__(learner)



from ._data_cleaner import DataCleaner
from ._dataframe_mapper import DataFrameMapper
from ._transformer import SafeOrdinalEncoder, FeatureImportancesSelectionTransformer
from ._transformer import general_preprocessor, general_estimator
